package edu.princeton.cs.schapire.lemonade;

import game.Game;
import game.Strategy;
import game.Utility;

import java.util.Random;

/**
 * Author: Robert Schapire (schapire@cs.princeton.edu)
 *
 * This strategy maintains estimates of the probability that each of
 * the other players will play each of the game positions using a
 * simple Laplace ("add one") smoothing rule.  Based on these
 * estimated probabilities, on each round, an action is chosen so as
 * to maximize the expected difference between the cumulative utility
 * gained by this player and the maximum cumulative utility of all the
 * other players.
 *
 **/

public class GreedyExpectedLaplaceStrategy implements Strategy{
  /**
   * 
   * @param g
   * @param seed
   * @param options
   */

    // defaults for parameters
    private double initCountParm = 1.0;
    private double oldActionWeightParm = 1.0;
    private double newActionWeightParm = 1.0;

    private Random random;
    private int[] cumUtil;
    private int[] util;
    private double probAction[][];
    private double actionCount[][];
    private int actionChoice[];

    private Utility utility;
    private int numPlayers;
    private int[] numActions;

    public GreedyExpectedLaplaceStrategy(Game g, long seed, String options) {
	parseOptions(options);

	utility = g.getUtility();
	numPlayers = g.getNumPlayers();
	numActions = new int[numPlayers];
	cumUtil = new int[numPlayers];
	util = new int[numPlayers];
	actionChoice = new int[numPlayers];

	probAction = new double[numPlayers][];
	actionCount = new double[numPlayers][];

	for (int i = 0; i < numPlayers; i++) {
	    numActions[i] = g.getNumActions(i);
	    probAction[i] = new double[numActions[i]];
	    actionCount[i] = new double[numActions[i]];
	    for (int a = 0; a < numActions[i]; a++)
		actionCount[i][a] = initCountParm;
	}
	random = new Random(seed);

    }

    private int chooseAction() {
	int best_action = 0;
	int num_best = 0;
	double best_exp_util = Double.NEGATIVE_INFINITY;

	for (int a = 0; a < numActions[0]; a++) {
	    actionChoice[0] = a;
	    double exp_util = expectedUtility(1);
	    if (exp_util > best_exp_util) {
		best_action = a;
		num_best = 1;
		best_exp_util = exp_util;
	    } else if (exp_util == best_exp_util
		       && random.nextInt(++num_best) == 0) {
		best_action = a;
	    }
	}
	return best_action;
    }
		
    double expectedUtility(int n) {
	if (n >= numPlayers) {
	    utility.getUtility(actionChoice, util);
	    int m = Integer.MIN_VALUE;
	    int u;
	    for (int i = 1; i < numPlayers; i++)
		if ((u = cumUtil[i] + util[i]) > m)
		    m = u;
	    return ((double) (cumUtil[0] + util[0] - m));
	}
	
	double s = 0.;
	for (int a = 0; a < numActions[n]; a++) {
	    if (probAction[n][a] > 0.) {
		actionChoice[n] = a;
		s += probAction[n][a] * expectedUtility(n+1);
	    }
	}
	return s;
    }

    public int getAction() {
	for (int i = 0; i < numPlayers; i++) {
	    double s = 0.0;
	    for (int a = 0; a < numActions[i]; a++)
		s += actionCount[i][a];
	    for (int a = 0; a < numActions[i]; a++)
		probAction[i][a] = actionCount[i][a] / s;
	}
	return chooseAction();
    }

    public void observeOutcome(int[] actions) {
	utility.getUtility(actions, util);

	for (int i = 0; i < numPlayers; i++) {
	    for (int a = 0; a < numActions[i]; a++)
		actionCount[i][a] *= oldActionWeightParm;
	    actionCount[i][actions[i]] += newActionWeightParm;
	    cumUtil[i] += util[i];
	}

    }
  
    private void parseOptions(String options) {
	String [] opts = options.split("\\s");
	double initCountParm = 0.;
	double oldActionWeightParm = 0.;
	double newActionWeightParm = 0.;

	if (options.trim().equals("")) // use defaults
	    return;

	try {
	    int i = 0;
	    initCountParm = Double.parseDouble(opts[i++]);
	    oldActionWeightParm = Double.parseDouble(opts[i++]);
	    newActionWeightParm = Double.parseDouble(opts[i++]);
	} catch (Exception e) {
	    System.err.println("bad arguments received -- reverting to defaults\n");
	    return;
	}

	this.initCountParm = initCountParm;
	this.oldActionWeightParm = oldActionWeightParm;
	this.newActionWeightParm = newActionWeightParm;  
    }

}
