package brownstrategy.agent;

import game.Game;

import java.util.Random;

import util.Constants;


/**
 * A strategy for a player. All players view the game as if
 * they are player 0. A class implementing the strategy
 * type should also implement a constructor of the form
 * MyClass(game.Game game, long randomSeed, String options).
 * 
 * The agent must return a distribution of actions.
 * 
 * @author eric
 *
 */
public abstract class NondeterministicAgent extends Agent {

	public Random random;


	public NondeterministicAgent(Game g, long seed, String options) {
		super(g, seed, options);
		random = new Random(seed);
	}


	public abstract void observeOutcome(int[] actions);


	/**
	 * Returns distribution of actions that should be probabilistically taken.
	 * Values should sum to 1, but they will be normalized by the action selector anyway.
	 * @return
	 */
	public abstract double[] getActionDistribution();	

	/**
	 * Returns one of the actions, randomly selected according to the agent's distribution.
	 */
	public int getAction() {

		double[] normalizedDistribution = normalize( getActionDistribution() );
		double[] cumulativeNormalizedDistribution = cumulativeSum( normalizedDistribution );

		//Choose random val between 0 and 1.
		double val = random.nextDouble();	

		for (int i=0; i<Constants.NUM_CLOCK_POSITIONS; i++) {
			if (val < cumulativeNormalizedDistribution[i]) {
				return i;
			}
		}

		System.out.println("Error: random value " + val + " >= anything in cumulativeNormalizedDistribution.");
		System.out.println("  We should never get here! Returning action 0 by default.");
		return 0;
	}



	/**
	 * Normalizes a distribution so that it sums to 1.
	 * @param distribution
	 * @return
	 */
	public double[] normalize( double[] distribution ) {
		double[] normDistribution = new double[distribution.length];
		double sum = 0;
		for (int i=0; i<distribution.length; i++) {
			sum += distribution[i];
		}

		/*
		 * added handling for sum is zero.  this shouldn't ever
		 * happen, but if it does it returns a unifrom distribution
		 * 
		 * -jberg
		 */
		if(sum == 0) {
			double uniform = 1.0/distribution.length;;
			for (int i=0; i<distribution.length; i++) {
				normDistribution[i] = uniform;
			}
		}
		else {
			for (int i=0; i<distribution.length; i++) {
				normDistribution[i] = distribution[i] / sum;
			}
		}
		return normDistribution;
	}
	
	/**
	 * Normalizes a distribution so that it sums to 1.
	 * @param distribution
	 * @return
	 */
	public double[] normalize(int[] distribution ) {
		double[] normDistribution = new double[distribution.length];
		double sum = 0;
		for (int i=0; i<distribution.length; i++) {
			sum += distribution[i];
		}

		/*
		 * added handling for sum is zero.  this shouldn't ever
		 * happen, but if it does it returns a unifrom distribution
		 * 
		 * -jberg
		 */
		if(sum == 0) {
			double uniform = 1.0/distribution.length;;
			for (int i=0; i<distribution.length; i++) {
				normDistribution[i] = uniform;
			}
		}
		else {
			for (int i=0; i<distribution.length; i++) {
				normDistribution[i] = distribution[i] / sum;
			}
		}
		return normDistribution;
	}



	/**
	 * Returns the cumulative sum of a distribution.
	 * @param distribution
	 * @return
	 */
	public double[] cumulativeSum( double[] distribution ) {
		double[] cumulativeSumDistribution = new double[distribution.length];

		cumulativeSumDistribution[0] = distribution[0];
		for (int i=1; i<distribution.length; i++) {
			cumulativeSumDistribution[i] = cumulativeSumDistribution[i-1] + distribution[i];
		}
		return cumulativeSumDistribution;
	}


}
