package edu.arizona.cs;


import game.Game;
import game.Strategy;
import game.Utility;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.util.Random;
import java.util.Vector;

/**
 * For this strategy we chose several features (building on efforts such as TeamUP)
 * with which we profile each other player in order to define a state-space. Since
 * these features are continuous, we discretized each player's 'state' based on a simple
 * clustering of profile data recorded from trials using last year's agents. The cluster
 * centers are then used as potential states for each opponent, and the cross-product of
 * these defines the state-space of the game. Rmax learning was used to explore this
 * state-space, with the reward being the payout from each turn, minus the
 * difference between the average payouts of the other two players. This strategy
 * uses the policy learned from that exploration.
 * 
 * The features used were:
 *   - stick/leadership: how stationary is an agent, a la TeamUP
 *   - follow: is an agent trying to sit directly across from another agent, also from TeamUP
 *   - fight: a measure of how much an agent will move close or stay close to another agent.
 *   - pattern: whether an agent is moving repeatedly through the same 2, 3 or 4 spots;
 *              ACT-R specifically was seen to do this.
 *   - prescience: a measure of whether an agent seems to know what another agent will do next,
 *                 as evidenced by using best response to the current move, rather than the
 *                 previous move.
 *   - For each of the above features, the max witnessed value was also tracked and used as a
 *        feature.
 *   - rewarddiff: the difference between the average payouts of the other two players.
 * 
 * There are several issues with this approach that may be explored in future work. Many of
 * these features are specific to the model used for agent behavior. For instance, the only
 * aspect of leadership considered here is how much an agent stays in place, but there could
 * be many other leading behaviors, such as moving in a pattern or the "sandwich offer" used
 * by RLcubed. A better approach for discretizing the state-space would also be desirable, such
 * as the Parti-game algorithm. The current approach requires a 'good' clustering, which can
 * be difficult to evaluate or achieve. 
 * 
 * @authors wrightjb, twalsh
 *
 */
public class Cactusade implements Strategy{
  /**
   * The action the player plays
   */
  protected Random r;
  protected int numActions,numBetterActions,betterAction=-1;

  protected double[][] clusters;
  protected int numClusters;
  protected int numStates;
  protected int state=-1,prev_state=-1;
  protected FinalAgent zero;
  protected FinalAgent one;
  protected FinalAgent two;
  protected String[] betterActions = {"stay put",
									"move randomly",
									"move across from current position",
									"move across from opponent one",
									"move across from opponent two",
									"sit on opponent one",
									"sit on opponent two",
									"move next to one on left",
									"move next to one on right",
									"move next to two on left",
									"move next to two on right"
};
  protected  BufferedWriter dataFile = null;
  protected BufferedReader tFile = null;
  protected int turncount=0;
  
//  double[][] q;
  protected double gamma=0.9;
  protected double alpha=0.1;
  protected double epsilon=0.1;
  protected Utility utility;
  protected int[] actionCounts;
  //int[] policy;
  

  public Cactusade(Game g, long seed, String options, String dataFileName, String tFileName) {
	  //11 Feature
	  clusters = new double[][]{
				{3.2617,4.4591,3.7523,4.6782,3.2215,4.3178,0.9061,0.9694,0.0089,0.0806,0.736},
				{5.9242,5.9676,5.741,5.8614,2.4007,3.9323,0.9907,0.9965,0.0012,0.0191,1.9437},
				{5.5132,5.7592,4.1224,4.8472,2.7716,3.9467,0.9234,0.9746,0.005,0.0425,1.096},
				{4.4942,5.5357,4.2411,5.1262,3.1217,4.2829,0.923,0.9915,0.2272,0.6114,1.127},
				{5.8059,5.9343,5.6144,5.7763,3.9529,4.7412,0.9766,0.9946,0.0026,0.0309,2.6806},
				{2.6759,2.9114,3.0966,3.2998,2.643,2.8972,0.6166,0.677,0.0013,0.0018,2.8114},
				{3.081,4.5054,3.5267,4.6814,3.4107,4.3922,0.5868,0.9422,0.0098,0.0858,0.6834},
				{5.543,5.8216,3.4092,4.7114,3.6735,4.3955,0.6382,0.9419,0.0036,0.0349,1.3636},
				{5.2744,5.3451,2.3624,3.3261,4.6147,4.7263,0.1186,0.4513,0.0021,0.0159,3.0354},
				{0.1276,0.1287,0.2721,0.2739,0.2162,0.2181,0,0,0,0,0.4288},
	  };
	  
	  setup(g,seed,options,dataFileName,tFileName);
  }
  
  public Cactusade(Game g, long seed, String options) {
	  this(g,seed,options,null,null);
  }
  
  protected void setup(Game g, long seed, String options, String dataFileName, String tFileName){
	  
		numClusters = clusters.length;
		numStates = numClusters*numClusters*numClusters;
		  
		this.numActions = g.getNumActions();
		this.utility = g.getUtility();
		r = new Random(seed);
		numBetterActions = betterActions.length;
		
		zero = new FinalAgent(12,"",g.getUtility());
		one = new FinalAgent(12,"",g.getUtility());
		two = new FinalAgent(12,"",g.getUtility());
		
		actionCounts = new int[numBetterActions];
		java.util.Arrays.fill(actionCounts, 0);
		
		readPolicy();
  }
  
  @Override
  public int getAction() {
	  
	  if (turncount < 1)
		  this.betterAction = 1;
	  else
		  this.betterAction = policy[state];
	  
	  turncount++;
	  actionCounts[betterAction]++;
    
	  int action=-1;
    switch(betterAction){
    case 0: // stay put
    	action = zero.getLastAction();
    	break;
    case 1: // move randomly
    	action = r.nextInt(this.numActions);
    	break;
    case 2: // move across from current position
    	action = getAcross(zero.getLastAction());
    	break;
    case 3: // move across from opponent one
    	action = getAcross(one.getLastAction());
    	break;
    case 4: // move across from opponent two
    	action = getAcross(two.getLastAction());
    	break;
    case 5: // sit on opponent one
    	action = one.getLastAction();
    	break;
    case 6: // sit on opponent two
    	action = two.getLastAction();
    	break;
    case 7: // move next to one on left
    	action = (one.getLastAction()+1)%12;
    	break;
    case 8: // move next to one on right
    	action = (one.getLastAction()-1)<0 ? 11 : (one.getLastAction()-1);
    	break;
    case 9: // move next to two on left
    	action = (two.getLastAction()+1)%12;
    	break;
    case 10: // move next to two on right
    	action = (two.getLastAction()-1)<0 ? 11 : (two.getLastAction()-1);
    	break;
    }
    
  	return action;
  }
  
  protected int getAcross(int a){
	  return (a+6)%12;
  }

  @Override
  public void observeOutcome(int[] actions) {
	  	zero.appendAction(actions[0]);
	  	one.appendAction(actions[1]);
	  	two.appendAction(actions[2]);
	  
		this.state = getState();
  }
  
  protected int getState(){
	  
	    double[][] stats = {zero.getStats(one, two),
	    					one.getStats(zero, two),
	    					two.getStats(zero, one)};
	    
	    int state = 0;
		int pnum = 0;
	    for(double[] player : stats){
		double min=Double.POSITIVE_INFINITY, temp;
		int argmini=-1, i=0;
			for (double[] cluster : clusters){
				temp=euc_distance(player, cluster);
				if(temp < min){
					min = temp;
					argmini = i;
				}
				i++;
			}
			state += argmini * Math.pow(clusters.length, pnum);
			pnum++;
	    
	    }
		return state;
  }
  
    protected double[] concatenate(double[] stats0,double[] stats1,double[] stats2){
		double[] stats = new double[stats0.length+stats1.length+stats2.length];
		System.arraycopy(stats0, 0, stats, 0,stats0.length);
		System.arraycopy(stats0, 0, stats, stats0.length,stats1.length);
		System.arraycopy(stats0, 0, stats, stats0.length+stats1.length,stats2.length);
		return stats;
	}
  
  protected double euc_distance(double[] stats,double[] cluster){
		double ss = 0;
		for( int i=0;i<cluster.length;i++){
			ss+= Math.pow(stats[i]-cluster[i],2);
		}
		ss = Math.sqrt(ss);
		return ss;
	}
  
  

  protected int[] policy = new int[numStates];
  protected String policyFileName = "policyFile.bin";  
    
  protected void readPolicy(){
	  try {
		InputStream is = Cactusade.class.getResourceAsStream(policyFileName);
		ObjectInputStream inputStream = new ObjectInputStream(is);
		policy = (int[]) inputStream.readObject();
	} catch (FileNotFoundException e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	} catch (IOException e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	} catch (ClassNotFoundException e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	}
  }
  
}



/**
 * @author wrightjb
 *
 */
class FinalAgent{
 int numActions;
 private int strat[];
 private Vector<Integer> sequence = new  Vector<Integer>();
 private static final double GAMMA = 0.75;
 private static final double RHO = 1;
 
 private double followIndex1 = 0, followIndex2 = 0;
 private double maxFollow = 0;
 private double fightIndex1 = 0, fightIndex2 = 0;
 private double maxFight = 0;
 private double stickIndex = 0;
 private double maxStick = 0;
 private double prescientIndex = 0;
 private double maxPrescient = 0;
 
 private double sumG = 1;
 
 private Utility utility;
 private double sumUtility = 0;
 
 public FinalAgent(int actions, String strategy, Utility utility){
  this.numActions = actions;
  this.utility = utility;
  strat = new int[numActions];
 }
 
 public void resetSequence(){
	 sequence = new  Vector<Integer>();
	 followIndex1 = 0;
	 followIndex2 = 0;
	 maxFollow = 0;
	 fightIndex1 = 0;
	 fightIndex2 = 0;
	 maxFight = 0;
	 stickIndex = 0;
	 maxStick = 0;
	 prescientIndex = 0;
	 maxPrescient = 0;
	 
	 sumG = 1;
	 
	 sumUtility=0;
 }
 
 public void appendAction(int myAction){ 
  strat[myAction]++;
  sequence.add(myAction);
 }
 
 public void updateSumUtility(int utility){
	 sumUtility+=utility;
 }
 
 public double getAvgScore(){
	 return sumUtility/(sequence.size()-1);
 }
 
 public int getLastAction(){
	 return sequence.lastElement();
 }
 
 protected void actionCorr(int myAction, int otherOpponent, int hisOwnAction){
 }
 
 
 private void updateSumG(){
	 sumG = sumG*GAMMA + 1;
 }
 
 private double updateDiscountedIndex(double index, double update){
	 update = Math.pow(update, RHO);
	 index = (index*(sumG-1) + update)/sumG;
	 return index;
 }
 
 protected int getNearness(int a, int b){
	 return Math.abs(6-Math.abs(a - b));
 }
 
 protected int getDistance(int a, int b){
	 return 6-Math.abs(6-Math.abs(a - b));
 }
 
 protected int getMaxUtility(int[] actions){
	 int possibleUtility;
	 int maxUtility = 0;
	 for(int i = 0; i<12; i++){
		 actions[0] = i;
		 possibleUtility = utility.getUtility(actions)[0];
		 if (possibleUtility > maxUtility)
			 maxUtility = possibleUtility;
	 }
	 return maxUtility;
 }
 
 protected Vector<Integer> getSequence(){
  return sequence;
 }

 protected int currentAction(){
  return sequence.lastElement();
 }
 
 public double[] getStats(FinalAgent one, FinalAgent two){
	 return getStatsAllFeatures(one,two);
 }
 
 public double[] getStats9Features(FinalAgent one, FinalAgent two){
	 int size = sequence.size();
	 if (size > 1) {
		 updateSumG();
		 followIndex1 = updateDiscountedIndex(followIndex1, getDistance(sequence.get(size-2),one.sequence.get(size-1)));
		 if (followIndex1 > maxFollow) maxFollow = followIndex1;
		 followIndex2 = updateDiscountedIndex(followIndex2, getDistance(sequence.get(size-2),two.sequence.get(size-1)));
		 if (followIndex2 > maxFollow) maxFollow = followIndex2;
		 
		 fightIndex1 = updateDiscountedIndex(fightIndex1,getNearness(sequence.get(size-2), one.sequence.get(size-1)));
		 if (fightIndex1 > maxFight) maxFight = fightIndex1;
		 fightIndex2 = updateDiscountedIndex(fightIndex2,getNearness(sequence.get(size-2), two.sequence.get(size-1)));
		 if (fightIndex2 > maxFight) maxFight = fightIndex2;
		 
		 stickIndex = updateDiscountedIndex(stickIndex, getNearness(sequence.get(size-1),sequence.get(size-2)));
		 if (stickIndex > maxStick) maxStick = stickIndex;
		 
		 int[] actions = new int[] {sequence.get(size-1),one.sequence.get(size-1),two.sequence.get(size-1)};
		 prescientIndex = updateDiscountedIndex(prescientIndex,utility.getUtility(actions)[0] == getMaxUtility(actions) ? 1 : 0);
		 if (prescientIndex > maxPrescient) maxPrescient = prescientIndex;
	 }
	 
	 double rewardDiff = Math.abs(one.getAvgScore() - two.getAvgScore());
	 
	 double[] stats = {stickIndex, maxStick,
			           Math.max(followIndex1,followIndex2), maxFollow,
			           Math.max(fightIndex1,fightIndex2), maxFight,
			           prescientIndex, maxPrescient,
			           Double.isNaN(rewardDiff) ? 0 : rewardDiff}; 
	 
	 return stats;
 }
	
 public double[] getStatsAllFeatures(FinalAgent one, FinalAgent two){
	 int size = sequence.size();
	 if (size > 1) {
		 updateSumG();
		 followIndex1 = updateDiscountedIndex(followIndex1, getDistance(sequence.get(size-2),one.sequence.get(size-1)));
		 if (followIndex1 > maxFollow) maxFollow = followIndex1;
		 followIndex2 = updateDiscountedIndex(followIndex2, getDistance(sequence.get(size-2),two.sequence.get(size-1)));
		 if (followIndex2 > maxFollow) maxFollow = followIndex2;
		 
		 fightIndex1 = updateDiscountedIndex(fightIndex1,getNearness(sequence.get(size-2), one.sequence.get(size-1)));
		 if (fightIndex1 > maxFight) maxFight = fightIndex1;
		 fightIndex2 = updateDiscountedIndex(fightIndex2,getNearness(sequence.get(size-2), two.sequence.get(size-1)));
		 if (fightIndex2 > maxFight) maxFight = fightIndex2;
		 
		 stickIndex = updateDiscountedIndex(stickIndex, getNearness(sequence.get(size-1),sequence.get(size-2)));
		 if (stickIndex > maxStick) maxStick = stickIndex;
		 
		 int[] actions = new int[] {sequence.get(size-1),one.sequence.get(size-1),two.sequence.get(size-1)};
		 prescientIndex = updateDiscountedIndex(prescientIndex,utility.getUtility(actions)[0] == getMaxUtility(actions) ? 1 : 0);
		 if (prescientIndex > maxPrescient) maxPrescient = prescientIndex;
		 
	 }
	 
	 double rewardDiff = Math.abs(one.getAvgScore() - two.getAvgScore());
	 
	 double[] stats = {stickIndex, maxStick,
			           followIndex1,followIndex2, maxFollow,
			           fightIndex1,fightIndex2, maxFight,
			           prescientIndex, maxPrescient,
			           Double.isNaN(rewardDiff) ? 0 : rewardDiff}; 
	 
	 return stats;
 }
 
 public double[] getStats3Features(FinalAgent one, FinalAgent two){
	 int size = sequence.size();
	 if (size > 1) {
		 updateSumG();
		 followIndex1 = updateDiscountedIndex(followIndex1, getDistance(sequence.get(size-2),one.sequence.get(size-1)));
		 followIndex2 = updateDiscountedIndex(followIndex2, getDistance(sequence.get(size-2),two.sequence.get(size-1)));
		 
		 stickIndex = updateDiscountedIndex(stickIndex, getNearness(sequence.get(size-1),sequence.get(size-2)));
	 }
	 
	 double[] stats = {stickIndex, followIndex1,followIndex2}; 
	 
	 return stats;
 }
 
 public double[] getStats5Features(FinalAgent one, FinalAgent two){
	 int size = sequence.size();
	 if (size > 1) {
		 updateSumG();
		 followIndex1 = updateDiscountedIndex(followIndex1, getDistance(sequence.get(size-2),one.sequence.get(size-1)));
		 if (followIndex1 > maxFollow) maxFollow = followIndex1;
		 followIndex2 = updateDiscountedIndex(followIndex2, getDistance(sequence.get(size-2),two.sequence.get(size-1)));
		 if (followIndex2 > maxFollow) maxFollow = followIndex2;
		 
		 stickIndex = updateDiscountedIndex(stickIndex, getNearness(sequence.get(size-1),sequence.get(size-2)));
		 if (stickIndex > maxStick) maxStick = stickIndex;
	 }
	 
	 double[] stats = {stickIndex, maxStick,
			           followIndex1,followIndex2, maxFollow}; 
	 
	 return stats;
 }
 
 public double[] getStats5FeaturesPlusRewdiff(FinalAgent one, FinalAgent two){
	 int size = sequence.size();
	 if (size > 1) {
		 updateSumG();
		 followIndex1 = updateDiscountedIndex(followIndex1, getDistance(sequence.get(size-2),one.sequence.get(size-1)));
		 if (followIndex1 > maxFollow) maxFollow = followIndex1;
		 followIndex2 = updateDiscountedIndex(followIndex2, getDistance(sequence.get(size-2),two.sequence.get(size-1)));
		 if (followIndex2 > maxFollow) maxFollow = followIndex2;
		 
		 stickIndex = updateDiscountedIndex(stickIndex, getNearness(sequence.get(size-1),sequence.get(size-2)));
		 if (stickIndex > maxStick) maxStick = stickIndex;
	 }
	 
	 double rewardDiff = Math.abs(one.getAvgScore() - two.getAvgScore());
	 
	 double[] stats = {stickIndex, maxStick,
			           followIndex1,followIndex2, maxFollow,
			           Double.isNaN(rewardDiff) ? 0 : rewardDiff}; 
	 
	 return stats;
 }

}