package edu.cmu.psy.reitter.meta;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import game.Game;
import game.Strategy;
import game.Utility;

import util.Array;

/**
 * In the standard lemonade game with three agents and twelve location, 
 * if two agents remain fixed across from each other (i.e. locations 0 and 6, 1 and 7, etc.),
 * they will always divide 18 utility between themselves.  The lowest each gets is 6 and the greatest is 12.
 * The third opponent is relegated to 6 utility no matter where it lands.
 * Thus, the expected utility for each of the two cooperative players is 9, while the third opponent gets only 6.
 * 
 * This agent begins with a few random moves (up to INITIAL_RANDOM_TURNS).  
 * From there, it diligently tries to find a partner to which it can stand opposite.
 * With a certain probability (dependent on the turn number), it will either 
 *   remain in the same location (in case another agent is moving opposite to it), or
 *   attempt to move opposite from one of the other players.
 * 
 * If it does decide to move, it picks the best candidate (the player most likely to stay across from it)
 * by examining historical performance.
 * The agent will voluntarily leave a position in which it is across from another agent only in very special 
 * circumstances, but the other agent may leave its opposite position (i.e. the team-up configuration breaks).  
 * The number of turns that 
 * the agent has had success with the player is recorded.  When the configuration breaks, it
 * is considered either a "failure" or a "defection" if the number of turns of success are greater 
 * than DEFECTION_THRESHOLD.  
 * The limited circumstances where this agent breaks an opposition are when it is across from a 
 * fixed opponent, and the third player is either adjacent to this fixed opponent or across from it, 
 * sharing our spot.  If either of these cases occur after a short waiting period in the early part 
 * of the game, the agent will move adjacent to the fixed player. When the third player opts to 
 * become adjacent to the fixed player as well, it counts as a sandwich_attempt.  If this pattern is 
 * broken, that counts as a defection as well.  
 * The best candidate for the agent to attempt to create a new configuration with is 
 * determined by the function =min(FAILURE_WEIGHT*num_failures + DEFECTION_WEIGHT*num_defections - sandwich_attempts*SANDWICH_BONUS).
 * The philosophy behind the failure/defection distinction was that a defecting player might be
 * a more malicious opponent and should be avoided.  In testing the performance of various combinations
 * of DEFECTION_THRESHOLD, FAILURE_WEIGHT, DEFECTION_WEIGHT, and SANDWICH_BONUS, no significant trends were established,
 * so the usefulness of the function may not be worthy of its presence.
 * 
 * 
 * 
 * @author yaros/wunder
 *
 */

public class CoOpp implements Strategy {

	//number of turns that will have a strictly random move at the beginning of the game
    //(this can probably be deleted)
	private static final int INITIAL_RANDOM_TURNS = 0;
    
	//the threshold used to distinguish a broken cooperation configuration as
	//either a "failure" (fewer success turns) or a "defection" (more success turns)
	private static final int DEFECTION_THRESHOLD = 0; 

	//the weights used in the utility function to determine the best candidate to attempt to get across from
	// =min(FAILURE_WEIGHT*num_failures + DEFECTION_WEIGHT*num_defections)
	private static final int DEFECTION_WEIGHT = 2;
	private static final int FAILURE_WEIGHT = 0;
	
	//these are taken from the Game class provided in the constructor
	private int mNumPlayers;
	private int mNumActions;
	private Utility mUtility;

	//used to get random numbers, initialized with the provided seed
	private Random mRandom;

	//scores in the game by player - this is not actually used in any decision making
	private int[] mScores;
	
	//the turn number in the game
	private int mTurn = 0;
	
	//array of previous actions - gathered each turn from observeOutcome()
	private int[] mPrevActions;
	
	//probabilities of staying per turn 
	//in case there is no one across from us, we decide whether to stay in hopes that
	//someone gets across from us, otherwise we attempt to move across from the best candidate
    private static final double P_0_TO_5 = 0.4;
	private static final double P_6_TO_10 = 1;
	private static final double P_11_TO_20 = 0.95;
	private static final double P_21_TO_50 = 0.5;
	private static final double P_51_TO_100 = 0.85;
	
    //map of player (key) to number of turns that the player has been across from us (value)
	//a player is added to map only if they are across from us, otherwise they are removed
    //the counts (values) are used to determine if it is a failure or a defection when the 
	//  player is no longer across from us
	private Map<Integer,Integer> mTrackingMap = new HashMap<Integer,Integer>();
	
	//count of the failures per player - used in the utility to choose the best candidate to pursue
	private int[] mCoopFailures;
    //count of the defections per player - used in the utility to choose the best candidate to pursue
	private int[] mDefections;
	
	// sandwich variables
	private static int SANDWICH_BONUS = 2;
	private static int WAIT_TO_SANDWICH = 5;
	private static double SANDWICH_PROBABILITY = 0.25;
	private int SANDWICH_WAITING = 2;
	private boolean sandwich_game = false;
	private int[] sandwich_attempts;
	private int[] sandwich_defections;
	private int[] swCoopFailures;
	private int[] constant_actions; 
	private Map<Integer,Integer> swTrackingMap = new HashMap<Integer,Integer>();

	
	//constructor - options is ignored
	public CoOpp(Game g, long seed, String options) {
	    mNumPlayers = g.getNumPlayers();
	    mNumActions = g.getNumActions();
	    mPrevActions = new int[mNumPlayers];
	    mScores = new int[mNumPlayers];
	    mCoopFailures = new int[mNumPlayers];
	    mDefections = new int[mNumPlayers];
	    constant_actions = new int[mNumPlayers];
	    sandwich_attempts = new int[mNumPlayers];
	    swCoopFailures = new int[mNumPlayers];
	    sandwich_defections = new int[mNumPlayers];
	    mUtility = g.getUtility();
	    mRandom = new Random(seed);
	    if (mRandom.nextDouble()<SANDWICH_PROBABILITY)
	    	sandwich_game = true;
	//    if (sandwich_game)
	}
	
	@Override
	public int getAction() {
		//use a random move for the first few turns (usually set to be only the very first move)
		if(mTurn <= INITIAL_RANDOM_TURNS) {
			return mRandom.nextInt(mNumActions);
		}
		
		//check if we are across from anyone
		for(int i=1; i<mNumPlayers; i++) {
			if(   mPrevActions[0] == ((mPrevActions[i] + mNumActions/2) % mNumActions)
			   || mPrevActions[0] == ((mPrevActions[i] + (mNumActions+1)/2) % mNumActions)) {
				//we are across from player i
				if(mTrackingMap.containsKey(i)) {
					//player i is in the tracking map - we have been successful for a while
					//increment the number of success turns (map value)
					mTrackingMap.put(i, mTrackingMap.get(i)+1);
				} else {
					//someone just got across from us - cool
					mTrackingMap.put(i, 1);
				}
			} else if(mTrackingMap.containsKey(i)) {
				//we are not across from player i - but we had been last turn
				//the configuration has broken, so we need to count it as either a failure or a defection
				int successCount = mTrackingMap.get(i);
				if(successCount >= DEFECTION_THRESHOLD) {
					//greater than the threshold, so its a defection
					mDefections[i]++;
				} else {
					//less than the threshold, just a failure
					mCoopFailures[i]++;
				}
				//take the player out of the tracking map b/c its no longer across from us
				mTrackingMap.remove(i);
			}
		}
		
		// Watch for fixed players and whether the other player either tries to sandwich it 
		// or is on top of us.
		if (mTurn>2&&((fixedPlayer(1)&&!acrossPlayer(0,2))||(fixedPlayer(2)&&!acrossPlayer(0,1)))) {
			for(int i=1; i<mNumPlayers; i++) {
				// check if one player is making half a sandwich with a fixed player. 
				if (sandwichPlayer(i))
					sandwich_attempts[i]++;
				// or if we are successful at sandwiching.
				if (sandwichPlayer(i)) {
					if(swTrackingMap.containsKey(i)) {
						swTrackingMap.put(i, swTrackingMap.get(i)+1);
					} else {
						swTrackingMap.put(i, 1);
					}
				} else if(swTrackingMap.containsKey(i)) {
					int successCount = swTrackingMap.get(i);
					if(successCount >= DEFECTION_THRESHOLD) {
						sandwich_defections[i]++;
					} else {
						swCoopFailures[i]++;
					}
					mTrackingMap.remove(i);
				}
			}
//			if(!swTrackingMap.isEmpty()) {
				//excellent - looks like cooperation
				if (fullSandwichPlayer(1)||fullSandwichPlayer(2))
					return mPrevActions[0];

				// we are almost in the cooperative outcome, but one of the players has to switch to the other side (50/50).
				if ((fixedPlayer(1)&&adjacentPlayer(0,1)&&mPrevActions[0]==mPrevActions[2])||(fixedPlayer(2)&&adjacentPlayer(0,2)&&mPrevActions[0]==mPrevActions[1])) {
					if (fixedPlayer(1)) {
						return (mNumActions+mPrevActions[1]+1-2*(mRandom.nextInt(2)))%mNumActions;
					}
					else {
						return (mNumActions+mPrevActions[2]+1-2*(mRandom.nextInt(2)))%mNumActions;
					}
				}	
				// get in sandwich position if the other player is there also.
				if(!swTrackingMap.isEmpty()) {
					if (fixedPlayer(1))
						return (mNumActions+2*mPrevActions[1]-mPrevActions[2])%mNumActions;
					if (fixedPlayer(2))
						return (mNumActions+2*mPrevActions[2]-mPrevActions[1])%mNumActions;
				}
			// Now for the tricky part: if we are facing a constant player, should we try to sandwich unilaterally, 
			// and how long should we wait?
			// this option will only be performed with some probability.
			if (!fullSandwichPlayer(1)&&!fullSandwichPlayer(2)&&sandwich_game&&SANDWICH_WAITING>=0&&mTurn>WAIT_TO_SANDWICH) {
				if (fixedPlayer(1))
					mDefections[2]++;
				if (fixedPlayer(2))
					mDefections[1]++;
				SANDWICH_WAITING--;
			}
			
			if (sandwich_game&&fixedPlayer(1)&&acrossPlayer(1,2)&&SANDWICH_WAITING>=0&&mTurn>WAIT_TO_SANDWICH) {
				if (getClockwiseDist((mPrevActions[1]+1)%mNumActions,mPrevActions[1],mPrevActions[2])+
					getCounterclockwiseDist((mPrevActions[1]+1)%mNumActions,mPrevActions[1],mPrevActions[2])>=mNumActions/2) {
					return (mNumActions+mPrevActions[1]+1)%mNumActions;//-2*(r.nextInt(2)))%numActions;
				}
				else {
					return (mNumActions+mPrevActions[1]-1)%mNumActions;//+2*(r.nextInt(2)))%numActions;
				}
			}
			if (sandwich_game&&fixedPlayer(2)&&acrossPlayer(1,2)&&SANDWICH_WAITING>=0&&mTurn>WAIT_TO_SANDWICH) {
				if (getClockwiseDist((mPrevActions[2]+1)%mNumActions,mPrevActions[2],mPrevActions[1])+
					getCounterclockwiseDist((mPrevActions[2]+1)%mNumActions,mPrevActions[2],mPrevActions[1])>=mNumActions/2) {
					return (mNumActions+mPrevActions[2]+1)%mNumActions;//-2*(r.nextInt(2)))%numActions;
				}
				else {
					return (mNumActions+mPrevActions[2]-1)%mNumActions;//+2*(r.nextInt(2)))%numActions;
				}
			}
		}
		
		if(!mTrackingMap.isEmpty()) {
			//excellent - looks like cooperation b/c someone is across from us
			//so lets return our previous move
			return mPrevActions[0];
		}
		
		//so no one is across from us now and we want to get across from someone
		//maybe they are getting across from us so we should stay 
		// or maybe they are staying and we should get across from them
		//let's flip a coin and see if we should move
		if(mTurn < 5 && mRandom.nextDouble() < P_0_TO_5) {
			//coin says stay put
			return mPrevActions[0];
		} else if(5 <= mTurn && mTurn < 10 && mRandom.nextDouble() < P_6_TO_10) {
			//coin says stay put
			return mPrevActions[0];
		} else if(10 <= mTurn && mTurn < 20 && mRandom.nextDouble() < P_11_TO_20) {
			//coin says stay put
			return mPrevActions[0];
		} else if(20 <= mTurn && mTurn < 50 && mRandom.nextDouble() < P_21_TO_50) {
			//coin says stay put
			return mPrevActions[0];
		} else if(50 <= mTurn && mRandom.nextDouble() < P_51_TO_100) {
			//coin says stay put
			return mPrevActions[0];
		}

		//coin says lets try to move across from someone
		//we will go for the player that we have had the fewest problems with in the past
		//there might be multiple good opponents, so lets keep them in a set and replace the set
		// if we find an opponenet with a higher score
		Set<Integer> bestCandidateSet = new HashSet<Integer>(); 
		//put the first guy in as the best candidate - we will replace him if there is someone better
		bestCandidateSet.add(1);
		int bestBetrayalScore = mCoopFailures[1] * FAILURE_WEIGHT + mDefections[1] * DEFECTION_WEIGHT - sandwich_attempts[1] * SANDWICH_BONUS;
		for(int i=2; i<mNumPlayers; i++) {
			//get the score based on our best candidate utility function
			int betrayalScore = mCoopFailures[i] * FAILURE_WEIGHT + mDefections[i] * DEFECTION_WEIGHT - sandwich_attempts[i] * SANDWICH_BONUS;
			if (betrayalScore < bestBetrayalScore) {
				//better than the previous best candidate, replace the set
				bestCandidateSet.clear();
				bestCandidateSet.add(i);
				bestBetrayalScore = betrayalScore;
			} else if (betrayalScore == bestBetrayalScore) {
				//tied with the best, add to the set
				bestCandidateSet.add(i);
			}
		}
		
		//pick one of the best candidates from our at random 
		//(there is probably only one, but a tie is possible)
		int candidate = (Integer)bestCandidateSet.toArray()[mRandom.nextInt(bestCandidateSet.size())];
		
		//add our best candidate to our tracking map with success turns equal to zero
		//also, add any of the other players at the same location as tracking candidates (if any exist)
		for(int i=1; i<mNumPlayers; i++) {
			if(mPrevActions[i] == mPrevActions[candidate]) {
				mTrackingMap.put(i,0);
			}
		}
		
		//everything is setup, now lets return the action across from our best candidate
		if(mNumActions / 2 != (mNumActions + 1) / 2) {
			//the number of actions is odd, pick one of the "across" actions at random
			return (int)((mPrevActions[candidate] + (mNumActions + mRandom.nextDouble() + .5) / 2) % mNumActions); 
		} else {
			return (mPrevActions[candidate] + mNumActions / 2) % mNumActions; 
		}
	}
	
	@Override
	public void observeOutcome(int[] actions) {
		//record the previous actions
		for(int i = 0; i<actions.length; i++){
			if (mTurn>0&&mPrevActions[i]==actions[i])
				constant_actions[i]++;
			mPrevActions[i] = actions[i];
		}
		//update the scores
		int[] util = mUtility.getUtility(actions);
		for(int i=0; i<util.length; i++) {
			mScores[i] += util[i];
		}
		//increment the turns counter
		mTurn++;
	}
	
	private boolean acrossPlayer(int p, int q) {
		return (mNumActions+mPrevActions[p]-mPrevActions[q])%mNumActions==mNumActions/2;
	}
	
	private boolean adjacentPlayer(int p, int q) {
		return (mNumActions+mPrevActions[p]-mPrevActions[q])%mNumActions==1||(mNumActions+mPrevActions[q]-mPrevActions[p])%mNumActions==1;
	}
	 
	private boolean fixedPlayer(int p) {
		return constant_actions[p]>=mTurn-1;
	}
	
	private boolean fullSandwichPlayer(int p) {
		return (adjacentPlayer(p,3-p)&&fixedPlayer(3-p)&&mPrevActions[0]!=mPrevActions[p]&&adjacentPlayer(0,3-p)); 
	}
	
	private boolean sandwichPlayer(int p) {
		return (adjacentPlayer(p,3-p)&&fixedPlayer(3-p)&&mPrevActions[0]!=mPrevActions[p]); //adjacentPlayer(0,3-p)&&
	}
	
	  
	private int getClockwiseDist(int p, int q, int r) {
		  if (p > q && p > r) {
			  if (q >= r) {
				  return (mNumActions - p) + r;
			  }
			  else {
				  return (mNumActions - p) + q;
			  }
		  }
		  else {
			  if (q > p && q > r) {
				  if (p >= r) 
					  return (q - p);
				  else return (r - p);
			  }
			  else {
				  if (r > p && r > q) {
				    if (p >= q) 
						  return (r - p);
					  else return (q - p);
				  }
				  else {
					  if (p==q && p==r)
						  return mNumActions;
					  if (p==q) 
						  return (mNumActions - p) + r; 
					  if (p==r) 
						  return (mNumActions - p) + q;
					  if (r==q) 
						  return (r - p);
				  }
			  }
		  }
		  return mNumActions;
	  }
	  
	private int getCounterclockwiseDist(int p, int q, int r) {
		  int[] flippedPos = new int[3];
		  flippedPos[0] = mNumActions - 1 - p;
		  flippedPos[1] = mNumActions - 1 - q;
		  flippedPos[2] = mNumActions - 1 - r;
		  
		  if (flippedPos[0] > flippedPos[1] && flippedPos[0] > flippedPos[2]) {
			  if (flippedPos[1] >= flippedPos[2]) {
				  return (mNumActions - flippedPos[0]) + flippedPos[2];
			  }
			  else {
				  return (mNumActions - flippedPos[0]) + flippedPos[1];
			  }
		  }
		  else {
			  if (flippedPos[1] > flippedPos[0] && flippedPos[1] > flippedPos[2]) {
				  if (flippedPos[0] >= flippedPos[2]) 
					  return (flippedPos[1] - flippedPos[0]);
				  else return (flippedPos[2] - flippedPos[0]);
			  }
			  else {
				  if (flippedPos[2] > flippedPos[0] && flippedPos[2] > flippedPos[1]) {
				    if (flippedPos[0] >= flippedPos[1]) 
						  return (flippedPos[2] - flippedPos[0]);
					  else return (flippedPos[1] - flippedPos[0]);
				  }
				  else {
					  if (p==q && p==r)
						  return mNumActions;
					  if (p==q) 
						  return (mNumActions - flippedPos[0]) + flippedPos[2]; 
					  if (p==r) 
						  return (mNumActions - flippedPos[0]) + flippedPos[1];
					  if (r==q) 
						  return (flippedPos[2] - flippedPos[0]);
				  }
			  }
		  }
		  return mNumActions;
	  }
	  
}


