package edu.rutgers.rl3;

import java.util.HashMap;
import java.util.Vector;
import java.util.LinkedList;
import java.util.HashSet;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.text.*;
import util.Array;

import game.Game;
import game.Strategy;
import game.Utility;

import util.Array;

/**
 * COLLABORATOR
 * This lemonade stand agent works on the principle of fast action 
 * and reciprocation.  It starts by quickly moving across from 
 * a chosen player, unless the current configuration gives a utility 
 * of 9 or more.  If the population seems to move frequently, the agent 
 * slows down to wait for others to partner up.  
 * The key motivation behind initiating the coalition is that others 
 * will tend to wait to observe the population before deciding on a 
 * course of action, which gives movers better odds for ending in 
 * an across coalition.
 * 
 * We recognize that moving can make us a target for the third agent. 
 * We take the assumption that this "sucker" will not punish us for 
 * no reason, but instead may attempt to coax us into a partnership 
 * with itself.  As such the Collaborator is open to such a courtship 
 * if it is done correctly.  Specifically, there needs to be a mix of 
 * offers and punishment, roughly in the same proportion. 
 * It looks for the following patterns (assuming we are at 6 and our 
 * current partner at 12):
 * -alternating 6 (punish) and somewhere between 9 and 12, or 12 and 3 (reward offer)
 * -double alternating 6 and somewhere between 9 and 12, or 12 and 3 
 * 		(which means 2 on, 2 off)
 * -some relative ratio over 12 turns that matches at least a third 
 * 		punishment, and a third in some consistent offer 
 * 
 * The punishment will give us an incentive to join the player, 
 * and the need for consistency is to know which position to move into.
 * 
 * Are we taking a risk that our old partner will now come after us? 
 * Yes, but we will be open to cooperating with them again if they choose to.
 * We are taking the bait under the assumption that some players 
 * will in fact not know how to respond or at least will not be out for revenge. 
 * It is these players who we are in fact aiming to benefit from. 
 * 
 * In a similar way, if we are in the sucker position, we might as well 
 * try to convince others to join up with us. 
 * We attempt the double switching, 2 on, 2 off, in order to give others 
 * a turn to jump across from us.  If this technique hasn't worked 
 * we switch targets partway through the game.  
 * 
 * It would be nice to have an agent 
 * that is more proactive in exploring which combination of moves 
 * will be most convincing, but it is difficult to do so in a consistent way. 
 * For now the most experimental aspect is how we identify 
 * a better partner. 
 * 
 * @author wunder/kaisers/littman/yaros
 *
 */

public class Collaborator implements Strategy {

	//number of turns that will have a strictly random move at the beginning of the game
    //(this can probably be deleted)
	private static final int INITIAL_RANDOM_TURNS = 0;

    private final static int SATISFACTORY_UTILITY = 9;
    
	//the threshold used to distinguish a broken cooperation configuration as
	//either a "failure" (fewer success turns) or a "defection" (more success turns)
	private static final int DEFECTION_THRESHOLD = 0; 

	//the weights used in the utility function to determine the best candidate to attempt to get across from
	// =min(FAILURE_WEIGHT*num_failures + DEFECTION_WEIGHT*num_defections)
	private static final int DEFECTION_WEIGHT = 2;
	private static final int FAILURE_WEIGHT = 0;
	
	//these are taken from the Game class provided in the constructor
	private int numPlayers;
	private int numActions;
	private Utility mUtility;

	//used to get random numbers, initialized with the provided seed
	private Random mRandom;

	//scores in the game by player - this is not actually used in any decision making
	private int[] mScores;
	
	//the turn number in the game
	private int turn = 0;
	
	//array of previous actions - gathered each turn from observeOutcome()
	private int[] prevActions;

	private int[][] allActions;
	private Vector<int[]> mActionHistory; // a list of past joint actions
	   
    private int mCollaborator = -1; // stores the current collaborator

	//probabilities of staying per turn 
	//in case there is no one across from us, we decide whether to stay in hopes that
	//someone gets across from us, otherwise we attempt to move across from the best candidate
    private static final double P_0_TO_5 = 0.0;
	private static final double P_6_TO_10 = 0.5;
	private static final double P_11_TO_20 = 0.5;
	private static final double P_21_TO_50 = 0.5;
	private static final double P_51_TO_100 = 0.9;
	
    //map of player (key) to number of turns that the player has been across from us (value)
	//a player is added to map only if they are across from us, otherwise they are removed
    //the counts (values) are used to determine if it is a failure or a defection when the 
	//  player is no longer across from us
	private Map<Integer,Integer> mTrackingMap = new HashMap<Integer,Integer>();
	
	//count of the failures per player - used in the utility to choose the best candidate to pursue
	private int[] mCoopFailures;
    //count of the defections per player - used in the utility to choose the best candidate to pursue
	private int[] mDefections;
	
	//tracking the frequencies of certain actions
	private int[] constant_actions; 
	private int[] across_actions; 
	private int[] other_actions; 
	
	// variables for dealing with the sucker situation
	private boolean irregularCarrot = true;
	private boolean look_for_carrot = false;
	private boolean carrot_and_sticking = false;
	private int carrotStartTurn = 0;
    
	private int DEBUG_LEVEL = 0;
	
	//constructor - options is ignored
	public Collaborator(Game g, long seed, String options) {
	    numPlayers = g.getNumPlayers();
	    numActions = g.getNumActions();
	    prevActions = new int[numPlayers];
	    mRandom = new Random(seed);
		prevActions[0] = mRandom.nextInt(numActions);
	    allActions = new int[100][numPlayers];
	    mScores = new int[numPlayers];
	    mCoopFailures = new int[numPlayers];
	    mDefections = new int[numPlayers];
	    constant_actions = new int[numPlayers];
	    across_actions = new int[numPlayers];
	    other_actions = new int[numPlayers];
	    mUtility = g.getUtility();
	    mActionHistory  = new Vector<int[]>();
	}
	
	@Override
	public int getAction() {
		//use a random move for the first few turns (usually set to be only the very first move)
		if(turn <= INITIAL_RANDOM_TURNS) {
			return mRandom.nextInt(numActions);
		}
		
		//check if we are across from anyone
		for(int i=1; i<numPlayers; i++) {
			if(   prevActions[0] == ((prevActions[i] + numActions/2) % numActions)) {
				if (prevActions[0]!=prevActions[3-i])			
					carrotStartTurn = 0;
				
				//we are across from player i
				if(mTrackingMap.containsKey(i)) {
					//player i is in the tracking map - we have been successful for a while
					//increment the number of success turns (map value)
					mTrackingMap.put(i, mTrackingMap.get(i)+1);
				} else {
					//someone just got across from us - cool
					mTrackingMap.put(i, 1);
				}
			} else if(mTrackingMap.containsKey(i)) {
				//we are not across from player i - but we had been last turn
				//the configuration has broken, so we need to count it as either a failure or a defection
				int successCount = mTrackingMap.get(i);
				if(successCount >= DEFECTION_THRESHOLD) {
					//greater than the threshold, so its a defection
					mDefections[i]++;
				} else {
					//less than the threshold, just a failure
					mCoopFailures[i]++;
				}
				//take the player out of the tracking map b/c its no longer across from us
				mTrackingMap.remove(i);
			}
		}
		
		// CARROT-AND-STICK section
		// We are currently across from someone but someone else is on our spot. 
		// We may need to rectify the situation.
		// If we were there first, we will keep our eyes open for cooperative offers.
		// If they were there first, we should start making an offer. Our standard 
		// offer will be to act at 9 o'clock to the constant's 12, and our 6.
		// We need to mix carrots and sticks together with some pattern. 
		// 1st attempt: a simple alternating pattern: play 6, play 9, and so on. 
		// At some point we may try something else if this is not working.
		// 2nd attempt: 2 moves carrot, 2 stick. play 6, play 6, play 9, play 9. 
		// 3rd attempt: up the offer. play 6, play 10, play 6, play 10. 
		// 4th attempt: Switch players and return to attempt 1. 
		// ** Keep this section SIMPLE! **
		if (turn>8&&(acrossPlayerThree(1,2)||look_for_carrot)) {
			// first, identify who is where. Are we across from P1 or P2?			
			int accP = -1;
			if (acrossPlayer(0,1))
				accP = 1;
			if (acrossPlayer(0,2))
				accP = 2;
			if (accP>0) {
				// if we have found a carrot, let's take it.
				if(look_for_carrot&&carrotFound(3-accP,accP)>=0) {
					look_for_carrot = false;
					return (carrotFound(3-accP,accP)+numActions/2)%numActions;
				}				
				else if (acrossFirst(0,3-accP,accP)) { // if we are across first, we should be looking for the carrot
					
					look_for_carrot = true;
					return prevActions[0];
				}
				
				// if we are across from one of the opposites, but were last to move, give the carrot 
				if (turn%2==0&&turn-carrotStartTurn>-1)
					return prevActions[0]; 
				else if (turn-carrotStartTurn>=25)
					return (prevActions[accP]-2+numActions)%numActions;
				else 
					return (prevActions[accP]-3+numActions)%numActions;
			}

			if (accP<0) { // We are the sucker, so we must begin carrot-and-sticking someone, starting with stick
				// keep track of the turn we started trying to carrot-and-stick, 
				// in case we want to switch to something else after a time
				if (carrotStartTurn<=0) 
					carrotStartTurn = turn;
				if (turn%2==0&&turn-carrotStartTurn>-1)//&&turn-carrotStartTurn<22)
					return prevActions[0];
				else if (turn-carrotStartTurn<50)
					return (prevActions[3-lastToMove()]+numActions/2)%numActions; 
				else 
					return (prevActions[lastToMove()]+numActions/2)%numActions; 
			}	
		}
		if (turn>8)
			if ((acrossPlayer(1,2)&&acrossPlayerThree(0,2)&&acrossFirst(0,1,2))||(acrossPlayer(1,2)&&acrossPlayerThree(0,1)&&acrossFirst(0,2,1)))
						look_for_carrot = true;
			
		// END CARROT-AND-STICK
				
		
		// make a note of who we might collaborate with
		if (turn<=1) {
			this.mCollaborator = mRandom.nextInt(2)+1; // index in action array: 1 or 2
			int[] utilities = this.mUtility.getUtility(prevActions);
			// 	select more satisfied collaborator
			if (utilities[1] > utilities[2]) this.mCollaborator = 1;
			else if (utilities[1] < utilities[2]) this.mCollaborator = 2;
		}
		
		if(!mTrackingMap.isEmpty()) {
			//excellent - looks like cooperation b/c someone is across from us
			//so lets return our previous move
			return prevActions[0];
		}
		
		// so no one is across from us now and we want to get across from someone
		// maybe they are getting across from us so we should stay 
		// or maybe they are staying and we should get across from them
		// let's flip a coin and see if we should move 
		// (P_0_TO_5=0 in early stages, later it is 0.5)
		if (turn<2 && mRandom.nextDouble() < P_0_TO_5) {
			return prevActions[0];
		} else if(turn < 5 && ((double)(across_actions[1]+across_actions[2])/(double)(2*turn)>0.7 && mRandom.nextDouble() < 1-P_0_TO_5)) {
			//coin says stay put
			return prevActions[0];
		} else if(turn < 5 && ((double)(across_actions[1]+across_actions[2])/(double)(2*turn)<=0.7 && mRandom.nextDouble() < P_0_TO_5)) {
			//coin says stay put
			return prevActions[0];
		} else if(5 < turn && turn < 10 && mRandom.nextDouble() < P_6_TO_10) {
			//coin says stay put
			return prevActions[0];
		} else if(10 <= turn && turn < 20 && mRandom.nextDouble() < P_11_TO_20) {
			//coin says stay put
			return prevActions[0];
		} else if(20 <= turn && turn < 50 && mRandom.nextDouble() < P_21_TO_50) {
			//coin says stay put
			return prevActions[0];
		} else if(50 <= turn && mRandom.nextDouble() < P_51_TO_100) {
			//coin says stay put
			return prevActions[0];
		}
		int[] util = mUtility.getUtility(prevActions);
		
		// For these initial turns, we should try to use a sophisticated approach when deciding when to move.
		if(turn < 6 && turn >= 1) {
			// we have observed some random behavior; penalize it
			
			if (other_actions[1]<other_actions[2])
				return (prevActions[1]+numActions/2)%2;
			if (other_actions[2]<other_actions[1])
				return (prevActions[2]+numActions/2)%2;
				
			if (true) {
		       switch(turn) {
	            case 0:
	                // the game starts, we don't know anything.
	                // Hence, act uniform random.
	                return mRandom.nextInt(numActions); // random move
	            case 1:
	                // We received the first move of every player. The best reply to this will be either Stay
	                // or Across, depending on the percentage of the population playing these two. Across has the
	                // additional benefit of signaling willingness to cooperate.
	                // Our best guess is, that the following move will be the same.
	                // Hence, choose a random player to cooperate with, since expected payoffs are equal.
	                this.mCollaborator = mRandom.nextInt(3-1)+1; // index in action array: 1 or 2
	                //prevActions = mActionHistory.lastElement();
	                int[] utilities = this.mUtility.getUtility(prevActions);
	                // select more satisfied collaborator
	                if (utilities[1] > utilities[2]) this.mCollaborator = 1;
	                else if (utilities[1] < utilities[2]) this.mCollaborator = 2;
	                // TODO: We could also choose the opponent who gains less from collaborating with us
	                // If utility is satisfactory, Stay
	                if (utilities[0] >= SATISFACTORY_UTILITY)
	                    return prevActions[0]; // Stay
	                // Move across from that player to establish a cooperation as soon as possible.
	                return this.getAcross(prevActions[mCollaborator]); // Across
	            case 2:
	            case 3:
	            case 4:
	                // We know two moves now. We can already classify into moving or not (mMoveCount).
	                // The moving can be classified as 'opposite' or not ('mAttemptedAcrossCount').
	                //                       Moving      |  Static
	                //        Opposite          | Unknown |
	                // -----------------------------------------------------
	                //  active collab (Am, Ao)  |   avoid | passive collab
	                //
	                // If we succeeded to collaborate last round, try to repeat that (i.e., with passive collab.).
	                if(this.isAcross(prevActions[0], prevActions[this.mCollaborator]) ) {
	                    return this.getAcross(prevActions[mCollaborator]);
	                }
	                int[] triggerActions = mActionHistory.get(mActionHistory.size()-2);
	                if(this.isAcross(triggerActions[0], prevActions[mCollaborator])) {
	                    // we have mutually signaled intentions for collaboration
	                    // best reply is mixing Stay and Across 50/50
	                    if (this.mRandom.nextBoolean()) {
	                        return prevActions[0]; // stay
	                    } else {
	                        return this.getAcross(prevActions[mCollaborator]);
	                    }
	                }
	                // If our opponent tried to cooperate with us (oAm)
	                if(this.isAcross(triggerActions[0], prevActions[this.getOpponent()])) {
	                    // he has signaled intentions for collaboration
	                    // best reply is to signal with Across
	                    // since his probability of Across is likely to be < 0.5
	                    return this.getAcross(prevActions[mCollaborator]);
	                }
	                // If our collaborator did not try to cooperate with us (a given since we check above),
	                // while the other one was static (played Stay) switch to other.
	                if(triggerActions[this.getOpponent()] == prevActions[this.getOpponent()]) {
	                    // select other opponent
	                    this.mCollaborator = this.getOpponent();
	                    return this.getAcross(prevActions[mCollaborator]);
	                }
	            default:
	            	return prevActions[0];
		       }
			}
		}

		
		// if we get this far, as a default lets try to move across from someone
		// we will go for the player that we have had the fewest problems with in the past
		// there might be multiple good opponents, so lets keep them in a set and replace the set
		// if we find an opponent with a higher score
		Set<Integer> bestCandidateSet = new HashSet<Integer>(); 
		//put the first guy in as the best candidate - we will replace him if there is someone better
		bestCandidateSet.add(1);
		int bestBetrayalScore = mCoopFailures[1] * FAILURE_WEIGHT + mDefections[1] * DEFECTION_WEIGHT;// - sandwich_attempts[1] * SANDWICH_BONUS;
		for(int i=2; i<numPlayers; i++) {
			//get the score based on our best candidate utility function
			int betrayalScore = mCoopFailures[i] * FAILURE_WEIGHT + mDefections[i] * DEFECTION_WEIGHT;// - sandwich_attempts[i] * SANDWICH_BONUS;
			if (betrayalScore < bestBetrayalScore) {
				//better than the previous best candidate, replace the set
				bestCandidateSet.clear();
				bestCandidateSet.add(i);
				bestBetrayalScore = betrayalScore;
			} else if (betrayalScore == bestBetrayalScore) {
				//tied with the best, add to the set
				bestCandidateSet.add(i);
			}
		}
		
		//pick one of the best candidates from our at random 
		//(there is probably only one, but a tie is possible)
		int candidate = (Integer)bestCandidateSet.toArray()[mRandom.nextInt(bestCandidateSet.size())];
		
		//add our best candidate to our tracking map with success turns equal to zero
		//also, add any of the other players at the same location as tracking candidates (if any exist)
		for(int i=1; i<numPlayers; i++) {
			if(prevActions[i] == prevActions[candidate]) {
				mTrackingMap.put(i,0);
			}
		}
		
		//everything is setup, now lets return the action across from our best candidate
		
		return (prevActions[candidate] + numActions / 2) % numActions; 
		
	}
	
	@Override
	public void observeOutcome(int[] actions) {
		//record the previous actions
		for(int i = 0; i<actions.length; i++){
			if (turn>1){
				if (prevActions[i]==actions[i]) {
					constant_actions[i]++;
				} else if ((numActions/2+prevActions[(i+1)%3])%numActions==actions[i]||(numActions/2+prevActions[(i+2)%3])%numActions==actions[i]) {
					across_actions[i]++;
				} else  {
					other_actions[i]++;
				}
			}
			prevActions[i] = actions[i];
			allActions[turn][i] = actions[i];
		}
		if (DEBUG_LEVEL>0)
			System.out.println(turn + " " + Array.arrayToNakedString(actions)+ " ");
		//update the scores
		int[] util = mUtility.getUtility(actions);
		for(int i=0; i<util.length; i++) {
			mScores[i] += util[i];
		}
		mActionHistory.add(actions.clone());
        
		//increment the turns counter
		turn++;
	}
	
	private int getDistance(int p, int q, int t) {
		return Math.min((allActions[t][p]-allActions[t][q]+numActions)%numActions,(allActions[t][q]-allActions[t][p]+numActions)%numActions);
	}
	
	private int getDistanceAct(int p, int q) {
		return Math.min((p-q+numActions)%numActions,(q-p+numActions)%numActions);
	}
	
	private boolean acrossPlayer(int p, int q) {
		return (numActions+prevActions[p]-prevActions[q])%numActions==numActions/2;
	}
	
	private boolean acrossPlayerThree(int p, int q) {
		return  (numActions+allActions[turn-1][p]-allActions[turn-1][q])%numActions==numActions/2 && 
				(numActions+allActions[turn-2][p]-allActions[turn-2][q])%numActions==numActions/2 && 
				(numActions+allActions[turn-3][p]-allActions[turn-3][q])%numActions==numActions/2;
	}
	
	private boolean adjacentPlayer(int p, int q) {
		return (numActions+prevActions[p]-prevActions[q])%numActions==1||(numActions+prevActions[q]-prevActions[p])%numActions==1;
	}
	 
	private boolean fixedPlayer(int p) {
		return constant_actions[p]>=turn-1;
	}
	
	private boolean fullSandwichPlayer(int p) {
		return (adjacentPlayer(p,3-p)&&fixedPlayer(3-p)&&prevActions[0]!=prevActions[p]&&adjacentPlayer(0,3-p)); 
	}
	
	private boolean sandwichPlayer(int p) {
		return (adjacentPlayer(p,3-p)&&fixedPlayer(3-p)&&prevActions[0]!=prevActions[p]); //adjacentPlayer(0,3-p)&&
	}
	
	// q has been offering a new spot somewhere closer to r. return the reflection of the action through the r-axis. 
	private int mirrorMove(int qAct, int rAct) {
		int d = getDistanceAct(qAct,rAct);
		if (((rAct+d)%numActions)==qAct)
			return (rAct-d+numActions)%numActions;
		else 
			return (rAct+d)%numActions;
	}
	
	// Check who was the last player to move
	private int lastToMove() {
		for (int i=turn-1; i>=0; i--) {
			if (allActions[i][1]!=allActions[turn-1][1])
				return 1;
			if (allActions[i][2]!=allActions[turn-1][2])
				return 2;
		}
		if (mRandom.nextDouble()<0.5)
			return 1;
		return 2;
	}
	
	// look for some carrot/stick pattern
	private int carrotFound(int q, int r) {
		if ((allActions[turn-3][q]==allActions[turn-1][q])&&(allActions[turn-1][q]==allActions[turn-1][0])&&getDistance(q,r,turn-2)<=3)
				return allActions[turn-2][q];
		if (turn > 7)
			if (allActions[turn-4][q]==allActions[turn-1][q]&&allActions[turn-5][q]==allActions[turn-1][q]&&allActions[turn-3][q]==allActions[turn-2][q]&&getDistance(q,r,turn-1)<=3)
				return allActions[turn-1][q];
		if (irregularCarrot&&turn>20)
			return scanForIrregularCarrotStick(12,q);
		else
			return -1;
	}
	
	// check if q has been mixing his carrots and sticks together in some irregular (but still noticable) pattern. 
	private int scanForIrregularCarrotStick(int numHistory, int q) {
		int[] carrotCount = new int[numActions];
		int stickCount = 0;
		for (int i=1; i<numHistory+1; i++) {
			if (getDistanceAct(allActions[turn-i][0],allActions[turn-i][q])<=1)
				stickCount++;
			if (getDistanceAct(allActions[turn-i][0],allActions[turn-i][q])>=3)
				carrotCount[allActions[turn-i][q]]++;
		}
		int maxCarrot = -1;
		int maxFreq = 0;
		for (int i=0; i<numActions; i++) {
			if (carrotCount[i]>maxFreq) {
				maxCarrot = i;
				maxFreq = carrotCount[i];
			}
		}
		if (maxCarrot<0)
			return -1;
		if (DEBUG_LEVEL>0)
			System.out.println((double)carrotCount[maxCarrot]/(double)numHistory>=0.35&&(double)stickCount/(double)numHistory>=0.35&&allActions[turn-1][q]==maxCarrot);
		if ((double)carrotCount[maxCarrot]/(double)numHistory>=0.35&&(double)stickCount/(double)numHistory>=0.35&&allActions[turn-1][q]==maxCarrot)
			return maxCarrot;
		return -1;
		
	}
	  
	// is player p across from q before r?
	private boolean acrossFirst(int p, int q, int r) {
		for (int i=turn-1; i>=0; i--) {
			if (allActions[i][p]!=allActions[turn-1][p])
				return false; //((allActions[i][q]!=allActions[turn-1][q]))
			if ((allActions[i][q]!=allActions[turn-1][q])&&allActions[i][r]==allActions[turn-1][r])
				return true;
		}
		return false;
	}
	
    private int getOpponent(){
        if (3==3) {
            if (this.mCollaborator==1) return 2;
            return 1;
        } else {
            System.out.println("Untested code");
            int nc = mRandom.nextInt(3-2)+1;
            if (this.mCollaborator <= nc) return nc+1; // skip collaborator
            return nc;
        }
    }
    
    private boolean isAcross(int a, int b) {
        return Math.abs(a - b) == numActions/2;
    }

    private int getAcross(int a) {
        return (a + numActions/2) % numActions;
    }
	
	private int getClockwiseDist(int p, int q, int r) {
		  if (p > q && p > r) {
			  if (q >= r) {
				  return (numActions - p) + r;
			  }
			  else {
				  return (numActions - p) + q;
			  }
		  }
		  else {
			  if (q > p && q > r) {
				  if (p >= r) 
					  return (q - p);
				  else return (r - p);
			  }
			  else {
				  if (r > p && r > q) {
				    if (p >= q) 
						  return (r - p);
					  else return (q - p);
				  }
				  else {
					  if (p==q && p==r)
						  return numActions;
					  if (p==q) 
						  return (numActions - p) + r; 
					  if (p==r) 
						  return (numActions - p) + q;
					  if (r==q) 
						  return (r - p);
				  }
			  }
		  }
		  return numActions;
	  }
	  
	private int getCounterclockwiseDist(int p, int q, int r) {
		  int[] flippedPos = new int[3];
		  flippedPos[0] = numActions - 1 - p;
		  flippedPos[1] = numActions - 1 - q;
		  flippedPos[2] = numActions - 1 - r;
		  
		  if (flippedPos[0] > flippedPos[1] && flippedPos[0] > flippedPos[2]) {
			  if (flippedPos[1] >= flippedPos[2]) {
				  return (numActions - flippedPos[0]) + flippedPos[2];
			  }
			  else {
				  return (numActions - flippedPos[0]) + flippedPos[1];
			  }
		  }
		  else {
			  if (flippedPos[1] > flippedPos[0] && flippedPos[1] > flippedPos[2]) {
				  if (flippedPos[0] >= flippedPos[2]) 
					  return (flippedPos[1] - flippedPos[0]);
				  else return (flippedPos[2] - flippedPos[0]);
			  }
			  else {
				  if (flippedPos[2] > flippedPos[0] && flippedPos[2] > flippedPos[1]) {
				    if (flippedPos[0] >= flippedPos[1]) 
						  return (flippedPos[2] - flippedPos[0]);
					  else return (flippedPos[1] - flippedPos[0]);
				  }
				  else {
					  if (p==q && p==r)
						  return numActions;
					  if (p==q) 
						  return (numActions - flippedPos[0]) + flippedPos[2]; 
					  if (p==r) 
						  return (numActions - flippedPos[0]) + flippedPos[1];
					  if (r==q) 
						  return (flippedPos[2] - flippedPos[0]);
				  }
			  }
		  }
		  return numActions;
	  }
	  
}


