package strategy;

import game.Game;
import game.Strategy;

import java.util.Random;
import java.util.Vector;
import lemonade.GeneralizedLemonadeUtility;



/**
 * Initial bot, evaluates all locations on the board and picks
 * the (or an equivalent) best spot and plays there for every round.
 * @author Tom Goff
 *
 */
public class BrownBot implements Strategy
{
	double[] locEvals;
	int bestSpot;
	Random r;
	GeneralizedLemonadeUtility utility; //stores the utility information about the island
	int lastcounter = 0;//count how many turns you've been the lowest scorer

	//used to keep track of this player's past moves, as well as opponents
	BrownBot[] opponents;
    Vector<Integer> moveHist;//tracks past moves for the player = new Vector();
    Vector<Integer> scoreHist;//tracks history of player scores in each round
	
	
	
    //Given two locations, return the shortest distance between them on the 12-spotted island
    static public int dist(int a, int b){
        int big = b, small = a;
        if(a > b){
            big = a;
            small = b;
        }
        return Math.min( big - small, 12 - big + small );
    }
	
    
	//evaluate the utility of a location given you can guess the next move of an opponent
	int guessQuality(int myAction, int oppAction){
	    int[] util = utility.getUtility(new int[]{myAction,oppAction,oppAction});
	    return util[0];
	}

	//estimate the best move if you can guess the next move of an opponent 
	private int guessBestAction(int opp1, int opp2){
	    int[] quality = new int[utility.getIslandSize()];
	    for(int i=0;i<quality.length;i++)
	    	quality[i]= (utility.getUtility(new int[]{i,opp1,opp2}))[0];
	    int maxQuality = quality[0];
	    for(int i=0;i<quality.length;i++)
	    	if (maxQuality<quality[i])
	    		maxQuality=quality[i];
	    int bestAction = 0;
	    for(int i=quality.length-1; i>=0; i--)//loop down to default to lower indexes
	    	if (quality[i]==maxQuality)
	    		bestAction = i;
	    return bestAction;
	}
	
	  
	//Evaluate spaces by their inherent density, and that of their neighbors
	public int evaluateLocations(GeneralizedLemonadeUtility utility){
		int[] density = utility.getIslandDensity(); //population at each location
		double[] locEvals = new double[ density.length ]; //stores evaluations, which reward being next to dense locations
		
		//evaluate each location by its own population and that of its neighbors
		for(int i=0; i < density.length; i++){
			//evaluation is density of the current location, plus reduced value of neighboring locations
			if(i == 0)//deal with wrap-around 
				locEvals[0] = density[0] + 0.5*(density[1] + density[density.length-1]) + 0.25*(density[2] + density[density.length-2]);
			else if( i == 1)
				locEvals[1] = density[1] + 0.5*(density[2] + density[0]) + 0.25*(density[3] + density[density.length-1]);
			else if( i == density.length-1)
				locEvals[i] = density[i] + 0.5*(density[0] + density[i-1]) + 0.25*(density[1] + density[i-2]);
			else if( i == density.length-2)
				locEvals[i] = density[i] + 0.5*(density[i+1] + density[i-1]) + 0.25*(density[0] + density[i-2]);	
			else//no need to worry about wrap-around
				locEvals[i] = density[i] + 0.5*(density[i+1] + density[i-1]) + 0.25*(density[i+2] + density[i-2]);
		}
		//with evaluations for each location.  Find the best spot, or pick one tied for best.
		double maxEval = 0;
		bestSpot = 0;
		for(int i=0; i < locEvals.length; i++)
			if( locEvals[i] > maxEval ){ 
				bestSpot = i;
				maxEval = locEvals[i];
			}
		return bestSpot;
	}
	  
	  
	  
	  
	/**
	* Constructor. The game is used to get the number of actions
	* and their respective populations.
	*
	* @param g
	* @param seed
	*/
	public BrownBot(Game g, long seed, String options) {
		this.r = new Random(seed + 8);
		this.lastcounter = 0;
		this.utility = (GeneralizedLemonadeUtility) g.getUtility();
		
		this.moveHist = new Vector<Integer>();
        this.scoreHist = new Vector<Integer>();
		
        this.opponents = new BrownBot[2];
        this.opponents[0] = new BrownBot();
        this.opponents[1] = new BrownBot();
        
		//evaluate each location on the island, rewarding both density and dense neighbors
		evaluateLocations( this.utility );

	}
	  
	//used to make models of opponent players, and track their moves over previous turns
	public BrownBot() {
		this.moveHist = new Vector<Integer>();
        this.scoreHist = new Vector<Integer>();
	}



	
	
	@Override
	public int getAction() {
		if( this.moveHist.size() < 5 ) //first moves. Stay on spot with best evaluation
            return evaluateLocations(this.utility);
		
		
		//check if you were the lowest scorer in the previous round...
		if( this.scoreHist.lastElement() <= this.opponents[0].scoreHist.lastElement() &&
			this.scoreHist.lastElement() <= this.opponents[0].scoreHist.lastElement())
			this.lastcounter++;
		else
			this.lastcounter = 0; //reset the count if you weren't in last place
		int lastMove = moveHist.lastElement();
		int lastA = opponents[0].moveHist.lastElement();
		int lastB = opponents[1].moveHist.lastElement();
		
		
		//a weird case... check and see if both opponents have been making EXACTLY
		//the same moves since the start of the game... if so, move randomly :P
		boolean copycats = true;
		for(int i=0; i < this.moveHist.size(); i++)
			if( opponents[0].moveHist.get(i) != opponents[1].moveHist.get(i))
				copycats = false;
		if( copycats && this.moveHist.size() > 10)
			return this.r.nextInt(12);//this takes advantage of 'bug' in the tournament
		
		
		
		//if you were the top scorer in the last round, don't move!
		if( this.scoreHist.lastElement() >= this.opponents[0].scoreHist.lastElement() && 
			this.scoreHist.lastElement() >= this.opponents[1].scoreHist.lastElement())
			return this.moveHist.lastElement();
		
		
		
		//test if you are getting pinched.  If so, move based on opponents last moves
		//if we're in a clear pinch between opponents... assume they'll stay where they are
		if( dist(lastMove, lastA) <= 2 && dist(lastMove, lastB) <= 2 && //pinch occurred...
			dist(lastMove, lastA) < dist(lastA, lastB) )
			return guessBestAction(lastA, lastB);
		
		
		
		//if you HAVE BEEN in last place for some threshold... consider moving
		if( this.lastcounter > 6){
			this.lastcounter = 0;//reset the counter
			int bestGuess = guessBestAction(lastA,lastB);//expected value of best guess
			int guessEval = utility.getUtility(new int[]{bestGuess,lastA,lastB})[0];
			
			if(guessEval > scoreHist.lastElement())//best counter-move, unless same (would have same score)
				return bestGuess;
			//else if( evaluateLocations(this.utility) != lastMove)//best spot, unless same...
			//	return evaluateLocations(this.utility);
			
			//go on offensive against whichever player is scoring better.  Hope to change game state
			if(this.opponents[0].scoreHist.lastElement() < this.opponents[1].scoreHist.lastElement())
				return lastB;
			return lastA;//if you seem hopelessly trapped... go on the offensive
		}
		
		//we're not clearly winning... not clearly losing... why not be greedy?
		return evaluateLocations(this.utility);
	}
	
	
	
	//int[3] actions holds the last moves of each player.  This player at actions[0]
	@Override
	public void observeOutcome(int[] actions){
		int[] scores = this.utility.getUtility(actions); //[mine, A's, B's]
	    
		this.moveHist.add( new Integer(actions[0]) );
	    this.scoreHist.add( new Integer(scores[0]) );
	      
	    this.opponents[0].moveHist.add( new Integer(actions[1]) );//opponent A
	    this.opponents[0].scoreHist.add( new Integer(scores[1]) );
	        
	    this.opponents[1].moveHist.add( new Integer(actions[2]) );//opponent B
	    this.opponents[1].scoreHist.add( new Integer(scores[2]) );
	}//end observeOutcome
}//end of BrownBot