package edu.gatech.macdermed;

import game.Game;
import game.Strategy;
import game.Utility;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.Random;

import util.Array;

import lemonade.LocationGameUtility;



/**
 * A strategy that tries to park across from someone else as quickly as possible
 * @author Liam
 *
 */
public class MatchMateAdv implements Strategy{
	private final int NUMPLAYERS;
	private final int NUMACTIONS;
  /**
   * The action the player plays
   */
  private Random r;
  private LocationGameUtility utility;  
  int myMove; 
  
  int roundNumber;
  int[] lastMoves;
  int[] moveCounts;

  enum modes {MATE, MATCH, SURRENDER };
  modes mode;
  private final int HOLD_TIME = 1;
  private final int SURRENDER_HORIZON = 12;
  private double SURRENDER_VALUE;// = 6.0;
  private final int MOVING_AVG_LENGTH = 6;
  private final double[] MOVE_POSITIONS;
  private final int TOTAL_CUSTOMERS;
  
  LinkedList<Integer> valueQueue;
  int queueValue;
  
  ArrayList<int[]> bestPairs;
  int mateMove;
  
  /**
   * Construct a constant strategy. The game is
   * used to get the number of actions, and the
   * seed is used to seed a pseudorandom number 
   * generator, which then is queried for the action 
   * to take.
   * 
   * @param g
   * @param seed 
   */
  public MatchMateAdv(Game g, long seed, String options) {
	NUMPLAYERS = g.getNumPlayers();
	NUMACTIONS = g.getNumActions();
   // int numActions = g.getNumActions();
    r = new Random(seed);
   this.utility = (LocationGameUtility) g.getUtility();
    moveCounts = new int[NUMPLAYERS];
    Arrays.fill(moveCounts, 0);
    valueQueue = new LinkedList<Integer>();
    for(int i=0; i<MOVING_AVG_LENGTH; i++) {
    	valueQueue.add(new Integer(0));
    }
    queueValue = 0;
    mode = modes.MATE;
    roundNumber = 0;
    // now to use utility information
    bestPairs = findBestPairs();
    bestPairs = findBestDifferences(bestPairs);
    int[] pair = bestPairs.get(r.nextInt(bestPairs.size()));
    int side = r.nextInt(2);
	myMove = pair[side];
    mateMove = pair[(side+1) % 2];
    SURRENDER_VALUE = getBestResponceUtilities(bestPairs.get(0))[2];
    int total = 0;
    for(int i=0; i<NUMACTIONS; i++) {
    	total += utility.getNumConsumers(i);
    }
    TOTAL_CUSTOMERS = total;
    
    MOVE_POSITIONS = new double[NUMACTIONS];
    double curPos = 0.0;
    for(int i=0; i<NUMACTIONS; i++) {
    	curPos += ((double)utility.getNumConsumers(i))/TOTAL_CUSTOMERS;
    	MOVE_POSITIONS[i] = curPos;
    }
    MOVE_POSITIONS[NUMACTIONS-1] = 1.0;
                   
    //System.out.println("**BEST PAIRS*** " + Arrays.deepToString(bestPairs.toArray()));
  }

@Override
  public int getAction() {
	roundNumber ++;
    return myMove;
  }

  @Override
  public void observeOutcome(int[] actions) {
	  // record if a player has moved or not
	  if(roundNumber != 1) {
		  for(int i=0; i<NUMPLAYERS; i++) {
			  if(actions[i] != lastMoves[i])
				  moveCounts[i] ++;
		  }
	  }
	  
	  // update moving average
	  int reward = utility.getUtility(actions)[0];
	  valueQueue.add(reward);
	  queueValue += reward;
	  queueValue -= valueQueue.removeFirst();
	  
	  // if doing poorly, surrender and revert to random strategy
	  if(roundNumber >= SURRENDER_HORIZON && ((double)queueValue)/MOVING_AVG_LENGTH < SURRENDER_VALUE && ! isOppositeSomeone(actions))
		  mode = modes.SURRENDER;
	  
	  switch (mode) {
	  	  case MATE:
		  case MATCH:	matchStrat(actions);  break;
		  case SURRENDER: 	myMove = getRandomMoveFromRelative(r.nextDouble());  break;
		  default:	throw new Error("Bad Mode in strategy MatchMate");
	  }
	  
	  lastMoves = actions;
  }
  
  // translates from discretized moves to 0-1 circle position
  double getRelativePosition(int move) {
	  return MOVE_POSITIONS[move];
  }
  
  // translates from 0-1 circle position to weighted position in the discretized game
  int getRandomMoveFromRelative(double position) {
	  double p = 0.0;
	  for(int i=0; i<NUMACTIONS; i++) {
		  double lowBound = p;
		  p = MOVE_POSITIONS[i];
		  if(p > position) {
			  int upP = (i+1) % NUMACTIONS;
			  if( position > (r.nextDouble()*(p - lowBound) + lowBound))
				  return upP;
			  else
				  return i;
		  }
	  }
	  return 0;
  }
  
  
  boolean isBest(int[] actions) {
	  int[] utils = utility.getUtility(actions);
	  for(int i=1; i<NUMPLAYERS; i++) {
		  if(utils[0] <= utils[i])
			  return false;
	  }
	  return true;
  }
  
  int mostSteadyOpp() {
	  int mostSteady = -1;
	  int value = Integer.MAX_VALUE;
	  int numEqual = 0;
	  for(int i=1; i<NUMPLAYERS; i++) {
		  if(moveCounts[i] < value) {
			  numEqual = 1;
			  mostSteady = i;
			  value = moveCounts[i];
		  } else if(moveCounts[i] == value) {
			  numEqual ++;
			  if(r.nextDouble() < 1.0/numEqual)
				  mostSteady = i;
		  }
	  }
	  return mostSteady;
  }
  
  public void mateStrat(int[] actions) {
	  boolean willStay = isOppositeMATE(actions) ||
	  					 isBest(actions) ||
	  					 roundNumber <= HOLD_TIME ||
	  					 r.nextDouble() > 1.0/2.0 ;
	  if(! willStay) {
		  mode = modes.MATCH;
		  matchStrat(actions);
	  }
  }
  
  public void matchStrat(int[] actions) {
	  boolean willStay = isOppositeSomeone(actions) ||
	  					 isBest(actions) ||
	  					 r.nextDouble() > 1.0/2.0 ;
	  if(! willStay)
		  	myMove = noisyOppositePosition(actions[mostSteadyOpp()]);
  }
  
  boolean isOppositeMATE(int[] actions) {
	  for(int i=1; i<NUMPLAYERS; i++)
			 if (actions[i] == this.mateMove)
				 return true;
	  return false;
  }
		
  boolean isOppositeSomeone(int[] actions) {
	  for(int i=1; i<NUMPLAYERS; i++)
		 if (actions[i] == noisyOppositePosition(actions[0]))
			 return true;
	  return false;
  }
  
  int noisyOppositePosition(int move) {
	double p = getRelativePosition(move)+0.5;
	if(p >= 1.0)
		p -= 1.0;
	return getRandomMoveFromRelative(p);
  }
  
  void printUtils() {
	  int total = 0;
	  for(int i=0; i<NUMACTIONS; i++) {
		  total += utility.getNumConsumers(i);
		  System.out.print(i + ":" + utility.getNumConsumers(i) + ", ");
	  }
	  System.out.println( " Total:" + total);
  }
  
  ArrayList<int[]> findBestPairs(){
	  int bestPairUtilValue = -1;
	  ArrayList<int[]> bestPairs = new ArrayList<int[]>();
	  for(int p1Move = 0; p1Move < NUMACTIONS; p1Move ++) {
		  for(int p2Move = p1Move+1; p2Move < NUMACTIONS; p2Move ++) {
			  int utilValue = getUtilityValue(p1Move, p2Move);
			  if(utilValue > bestPairUtilValue) {
				  bestPairUtilValue = utilValue;
				  bestPairs.clear();
			  }
			  if(utilValue == bestPairUtilValue) {
				  bestPairs.add(new int[]{p1Move, p2Move});
			  }
		  }
	  }
//	  printUtils();
//	  for(int[] pair : bestPairs) {
//		  System.out.println("pair: (" + pair[0] + ", " + pair[1] + ") value:" + bestPairUtilValue + " diff:" +
//				  getUtilityDifference(pair[0], pair[1]) + " 3rdMoves:" + findBestResponces(pair[0], pair[1]) + " AllScore(example) =" + Array.arrayToString(getBestResponceUtilities(pair[0], pair[1])));
//		  int[] upPair = new int[]{pair[0], pair[1], pair[0]};
//		  int[] downPair = new int[]{pair[0], pair[1], pair[1]};
//		  System.out.println("    Tied Up Pair " + Array.arrayToString(upPair) + " = " +  Array.arrayToString(utility.getUtility(upPair)) +
//				  "  Tied Down Pair " + Array.arrayToString(downPair) + ": " +  Array.arrayToString(utility.getUtility(downPair)));
//	  }
	  return bestPairs;
  }
  
  ArrayList<int[]> findBestDifferences(ArrayList<int[]> pairs) {
	  double bestDiff = Double.MAX_VALUE;
	  ArrayList<int[]> bestMoves = new ArrayList<int[]>();
	  for(int[] pair : pairs) {
		  double diff = getUtilityDifference(pair[0], pair[1]);
		  if(diff < bestDiff) {
			  bestDiff = diff;
			  bestMoves.clear();
		  }
		  if(diff == bestDiff) {
			  bestMoves.add(pair);
		  }
	  }
	  return bestMoves;
  }
  
  int getUtilityValue(int pos1, int pos2) {
	  int p3Move = getBest3rdResponce(pos1, pos2);
	  int[] move = {pos1, pos2, p3Move};
	  int[] utils = utility.getUtility(move);
	  return utils[0] + utils[1];
  }
  
  double getUtilityDifference(int p1Move, int p2Move) {
	  ArrayList<Integer> bestMoves = findBestResponces(p1Move, p2Move);
	  int sumP1 = 0;
	  int sumP2 = 0;
	  for(Integer p3Move : bestMoves) {
		  int[] u = utility.getUtility(new int[]{p1Move, p2Move, p3Move});
		  sumP1 += u[0];
		  sumP2 += u[1];
	  }
	  //assert (Math.abs(sumP1 - sumP2) % bestMoves.size() == 0);
	  return ((double) Math.abs(sumP1 - sumP2)) / bestMoves.size();
  }
  
  int[] getBestResponceUtilities(int[] pair) {
	  return getBestResponceUtilities(pair[0], pair[1]);
  } 
  int[] getBestResponceUtilities(int p1Move, int p2Move) {
	  int p3Move = getBest3rdResponce(p1Move, p2Move);
	  return utility.getUtility(new int[]{p1Move, p2Move, p3Move});
  }
  
  int getBest3rdResponce(int[] pair) {
	  return getBest3rdResponce(pair[0], pair[1]);
  }
  int getBest3rdResponce(int p1Move, int p2Move) {
	  int bestMove = -1;
	  int bestUtil = -1;
	  for(int p3Move = 0; p3Move < NUMACTIONS; p3Move ++) {
		  int[] move = {p1Move, p2Move, p3Move};
		  int p3Util = utility.getUtility(move)[2];
		  if(p3Util > bestUtil){
			  bestUtil = p3Util;
			  bestMove = p3Move;
		  }
	  }
	  return bestMove;
  }
  ArrayList<Integer> findBestResponces(int p1Move, int p2Move){
	  int bestUtil = -1;
	  ArrayList<Integer> bestMoves = new ArrayList<Integer>();
	  for(int p3Move = 0; p3Move < NUMACTIONS; p3Move ++) {
		  int[] move = {p1Move, p2Move, p3Move};
		  int p3Util = utility.getUtility(move)[2];  
		  if(p3Util > bestUtil) {
			  bestUtil = p3Util;
			  bestMoves.clear();
		  }
		  if(p3Util == bestUtil) {
			  bestMoves.add(new Integer(p3Move));
		  }
	  }
	  return bestMoves;
  }
  
}
