package au.edu.sydney;

import game.Game;
import game.Utility;

import java.util.Random;
import java.util.Arrays;

import com.yahoo.lsg.lifelong.GameInformation;
import com.yahoo.lsg.lifelong.LifelongStrategy;

/**
 * A collusion-seeking agent, with initial action choice adapted over games.
 *
 * Since the initial round's play in the LSG has been shown to be very important
 * to long- run success, we model the initial choices separately to the rest of
 * the game. Thus, we focus our super-game learning on initial play, and
 * subsequent adaptation during each game is left unaltered from game to game.
 * This design is also in part due to the success of simple strategies in the
 * 2011 tournament (with credit to the JerseyBeachcomber team and Michael
 * Bowling -- imitation is the sincerest form of flattery).
 *
 * In more detail:
 *
 * Initial opponent choices are modelled as k-level reasoning, with the
 * distribution over values of k adapted between games, as in Haruvy and Stahl
 * (GEB, 2012).
 *
 * During a game, if there is available an action that could lead to a coalition
 * forming, which is defined as an action profile that gives this agent receives
 * more that at least one of its opponents, then this action is selected. If
 * more than one such action exists, then the action that minimises the sum of
 * the square its opponents' immediate rewards is chosen, in order to minimise
 * the gain to any one opponent an for fairness and long-run stability reasons.
 * If no such coalition-initialising actions are available, the agent maximises
 * its own payoff.
 *
 * @author archiec
 */
public class BlackWattle implements LifelongStrategy {

    Utility utility;
    Random r;
    public int numActions;
    int nextAction;
    int iteration;
    double[] initialDist;
    double[][] sigma;
    double[][] kLevelStrategies;
    // Play with these values
    static final int K = 4; // number of levels (0,1,2...) 
    static final int D = 200; // number of samples to construct initial play and evaluate beliefs
    static final double TEMP = 0.008; // > 0 ; Boltzmann temperature parameter for k-level reasoning induction
    static final double LEARNING_RATE = 1.0; // \in (0,1]; Controls rate of initialDist updates 

    public BlackWattle(GameInformation info, long seed, String options) {
        r = new Random(seed);
        nextAction = -1;
        initialDist = new double[K];
        setIinitialDist(); // Lifetime initialisation
    }

    @Override
    public void newGame(Game g, int iterations) {
        this.numActions = g.getNumActions();
        this.utility = g.getUtility();

        iteration = 0;
        sigma = new double[K][numActions];
        kLevelStrategies = new double[K][numActions];
        kLevelReasoner(TEMP);
        initialAction();
    }
    
    @Override
    public void observeOutcome(int[] actions) {
        iteration++;
        if (iteration == 1) {
            updateInitialDist(actions);
        }
        setNextAction(actions);
    }
    
    @Override
    public int getAction() {
        return nextAction;
    }

    /**
     * Set initial distribution for the lifetime; hard-coded for uniform.
     *
     * Uniform. Can be tinkered with. 
     */
    private void setIinitialDist() {
        double sum = 0;
        for (int k = 0; k < K; k++) {
            initialDist[k] = 1.0 / K;
            sum += initialDist[k];
        }
        assert sum == 1;
    }

    /**
     * Return an initial action based on distribution over K.
     *
     * Takes D samples actions for opponents from initialDist, computes
     * bestReplies to each sampled profile, takes total bestReply frequencies as
     * a mixed strategy, sigma, and randomly chooses an action according to
     * sigma.
     *
     * Updates sigma, the jointProb of each action for each k, for
     * updateInitialDist.
     *
     */
    private void initialAction() {
        double[] marginalSigma = new double[numActions];
        for (int k = 0; k < K; k++) {
            for (int d = 0; d < D; d++) {
                int[] sampleActions = new int[3];
                sampleActions[0] = 0;
                for (int i = 1; i < sampleActions.length; i++) {
                    sampleActions[i] = randSample(kLevelStrategies[k]);
                }
                int[] bestReplies = getBestReplies(getPossibleUtilities(sampleActions));
                double[] dist = new double[bestReplies.length];
                for (int j = 0; j < bestReplies.length; j++) {
                    dist[j] = 1.0 / bestReplies.length;
                }
                int br = bestReplies[randSample(dist)];
                sigma[k][br] += 1.0 / D;
                marginalSigma[br] += 1.0 / (D * K);
            }
        }
        nextAction = getBestReplies(marginalSigma)[0];
//        nextAction = randSample(marginalSigma);
//        nextAction = r.nextInt(numActions);
    }

    /**
     * Recursively constructs the mixed strategy of a k-level reasoner in the
     * current game, beginning with a uniform belief, up to level K; with
     * trembles controlled by trembleTemp.
     *
     * @return
     */
    private void kLevelReasoner(double trembleTemp) {
        double[][] expectedReward = new double[K][numActions];
        // initial belief is uniform
        for (int j = 0; j < numActions; j++) {
            kLevelStrategies[0][j] = 1.0 / numActions;
        }
        for (int k = 1; k < K; k++) {
            // construct a sample of D joint actions based on kLevelStrategies[k-1]
            for (int d = 0; d < D; d++) {
                double[] belief = Arrays.copyOf(kLevelStrategies[k - 1], numActions);
                int[] possibleUtils = getPossibleUtilities(new int[]{d, randSample(belief), randSample(belief)});
                for (int j = 0; j < numActions; j++) {
                    expectedReward[k][j] += possibleUtils[j] / D;
                }
            }
            // Boltzmann distribution to add trembles to kLevelStrategy
            // if Temp -> 0 kLevelActions[k][bestReply(expectedReward[k])] = 1;
            double[] actionExponents = new double[numActions];
            double norm = 0.0;
            for (int j = 0; j < numActions; j++) {
                actionExponents[j] = Math.exp(expectedReward[k][j] / trembleTemp);
                norm += actionExponents[j];
            }

            for (int j = 0; j < numActions; j++) {
                kLevelStrategies[k][j] = actionExponents[j] / norm;
            }
        }
    }

    /**
     * Update kDistribution, based on initial play and previously calculated
     * sigma, using BNN-type dynamics
     */
    private void updateInitialDist(int[] actions) {
        // Evaluate each of the k levels' performance in light of actual "actions"
        double[] expectedRewardSigma = new double[K];
        double weightedAvgReward = 0.0;
        for (int k = 1; k < K; k++) {
            expectedRewardSigma[k] = 0;
            int[] testActions = actions;
            for (int j = 0; j < numActions; j++) {
                testActions[0] = j;
                expectedRewardSigma[k] += sigma[k][j] * (double) utility.getUtility(testActions)[0];
            }
            weightedAvgReward += expectedRewardSigma[k] * initialDist[k];
        }
        // If greater than averageReward, increase prob, if less than, decrease, 
        // according to BNN-type dynamics
        double[] excessReward = new double[K];
        double sumExcessRewards = 0.0;
        for (int k = 1; k < K; k++) {
            excessReward[k] = LEARNING_RATE * Math.max(expectedRewardSigma[k] - weightedAvgReward, 0);
            sumExcessRewards += excessReward[k];
        }
        for (int k = 1; k < K; k++) {
            initialDist[k] += excessReward[k] - initialDist[k] * sumExcessRewards;
        }
    }

    /**
     * Set the nextAction, based on current play.
     *
     * Simple heuristic. No planning; no forecasting
     *
     * @param actions
     */
    private void setNextAction(int[] actions) {
        int[] rewards = utility.getUtility(actions);
        // If doing worst, then change action.
        if (rewards[0] < rewards[1] && rewards[0] < rewards[2]) {
            // get all betterReplies
            int[] betterReplies = getBetterReplies(getPossibleUtilities(actions), actions);
            int[] rankedBetter = new int[betterReplies.length];
            int rankedBetterCounter = 0;
            for (int j = 0; j < betterReplies.length; j++) {
                int[] nextRewards = utility.getUtility(new int[]{j, actions[1], actions[2]});
                // Extract those that change the my ranking
                if (nextRewards[0] > nextRewards[1] || nextRewards[0] > nextRewards[2]) {
                    rankedBetter[rankedBetterCounter] = betterReplies[j];
                    rankedBetterCounter++;
                }
            }
            // get the rank improving betterreply that minSumSqrd others' rewards
            if (rankedBetterCounter > 0) {
                rankedBetter = Arrays.copyOf(rankedBetter, rankedBetterCounter);
                nextAction = getMinSsRewardAction(rankedBetter, actions);
            } else { // if none exist, do the best you can
                int[] bestReplies = getBestReplies(getPossibleUtilities(actions));
                nextAction = getMinSsRewardAction(bestReplies, actions);
            }
        }
    }

    /**
     * Returns the action that minimises the sum of squared opponent immediate
     * rewards, chosen from a set of feasible replies
     *
     * @param actions
     * @param feasibleResponses
     */
    public int getMinSsRewardAction(int[] feasibleResponses, int[] actions) {
        int action = feasibleResponses[0];
        if (feasibleResponses.length > 1) {
            int minSsRewards = 100;
            for (int j = 0; j < feasibleResponses.length; j++) {
                int ssOpponentRewards = (int) Math.pow(utility.getUtility(new int[]{feasibleResponses[j], actions[1], actions[2]})[1], 2);
                ssOpponentRewards += Math.pow(utility.getUtility(new int[]{feasibleResponses[j], actions[1], actions[2]})[2], 2);
                if (minSsRewards > ssOpponentRewards) {
                    minSsRewards = ssOpponentRewards;
                    action = feasibleResponses[j];
                }
            }
        }
        return action;
    }

    public int[] getPossibleUtilities(int[] actions) {
        int[] possibleUtilities = new int[numActions];
        for (int j = 0; j < possibleUtilities.length; j++) {
            actions[0] = j;
            possibleUtilities[j] = utility.getUtility(actions)[0];
        }
        return possibleUtilities;
    }

    /**
     * Returns a pure best reply given a vector of expected rewards, uniformly
     * randomises if there is a tie.
     *
     * E.g. for a bestReply to a pure action profile, try
     * bestReply(getPossibleUtilities(actions))
     *
     * @param action
     * @return
     */
    private int[] getBestReplies(double[] expRewards) {
        double bestReward = 0.0;
        int bestCounter = 0;
        int[] bestReplies = new int[expRewards.length];
        for (int j = 0; j < numActions; j++) {
            if (expRewards[j] > bestReward) {
                bestReward = expRewards[j];
                bestReplies[0] = j;
                bestCounter = 1;
            } else if (expRewards[j] == bestReward) {
                bestReplies[bestCounter] = j;
                bestCounter++;
            }
        }
        assert (bestCounter > 0);
        bestReplies = Arrays.copyOf(bestReplies, bestCounter);
        return bestReplies;
    }

    private int[] getBestReplies(int[] expRewards) {
        int bestReward = 0;
        int bestCounter = 0;
        int[] bestReplies = new int[expRewards.length];
        for (int j = 0; j < numActions; j++) {
            if (expRewards[j] > bestReward) {
                bestReward = expRewards[j];
                bestReplies[0] = j;
                bestCounter = 1;
            } else if (expRewards[j] == bestReward) {
                bestReplies[bestCounter] = j;
                bestCounter++;
            }
        }
        assert (bestCounter > 0);
        bestReplies = Arrays.copyOf(bestReplies, bestCounter);
        return bestReplies;
    }

    private int[] getBetterReplies(int[] expRewards, int[] actions) {
        int currentReward = expRewards[actions[0]];
        int[] betterReplies = new int[numActions];
        int betterCounter = 0;
        for (int j = 0; j < expRewards.length; j++) {
            if (expRewards[j] >= currentReward) {
                betterReplies[betterCounter] = j;
                betterCounter++;
            }
        }
        assert (betterCounter > 0);
        betterReplies = Arrays.copyOf(betterReplies, betterCounter);
        return betterReplies;
    }

    private int[] getBetterReplies(double[] expRewards, int[] actions) {
        double currentReward = expRewards[actions[0]];
        int[] betterReplies = new int[numActions];
        int betterCounter = 0;
        for (int j = 0; j < expRewards.length; j++) {
            if (expRewards[j] >= currentReward) {
                betterReplies[betterCounter] = j;
                betterCounter++;
            }
        }
        assert (betterCounter > 0);
        betterReplies = Arrays.copyOf(betterReplies, betterCounter);
        return betterReplies;
    }
    
    /**
     * Returns an index number randomly drawn from an arbitrary discrete
     * distribution.
     *
     * @param distribution
     * @return
     */
    private int randSample(double[] dist) {
        double rand = r.nextDouble();
        int j = 0;
        double nextThreshold = dist[j];
        while (rand > nextThreshold && j < dist.length - 1) {
            j++;
            nextThreshold += dist[j];
        }
        return j;
    }
    
//    /**
//     * Returns an array containing the PSNE consistent actions; returns "null"
//     * if no PSNE.
//     *
//     * @param actions
//     */
//    public int[] getEqResponses(int[] actions) {
//        int[] eqResponses = new int[numActions];
//        int eqCounter = 0;
//        int[] testAction = actions;
//        for (int j = 0; j < numActions; j++) {
//            testAction[0] = j;
//            if (isNash(testAction)) {
//                eqCounter++;
//                eqResponses[eqCounter] = j;
//            }
//        }
//        if (eqCounter > 0) {
//            eqResponses = Arrays.copyOf(eqResponses, eqCounter);
//            return eqResponses;
//        } else {
//            return null;
//        }
//    }
//    
//    /**
//     * Tests if an action profile is a one-shot PSNE, returns "false" if not.
//     *
//     * @param actions
//     */
//    public boolean isNash(int[] actions) {
//        boolean isNash = true;
//        int[] currentUtil = utility.getUtility(actions);
//        for (int i = 0; isNash && i < actions.length; i++) {
//            int[] testAction = actions;
//            for (int j = 0; isNash && j < numActions; j++) {
//                testAction[i] = j;
//                if (currentUtil[i] - utility.getUtility(testAction)[i] < 0) {
//                    isNash = false;
//                }
//            }
//        }
//        return isNash;
//    }
}