package edu.rutgers.rl3;


import game.Game;
import game.Strategy;

import java.util.Arrays;
import java.util.Random;
import java.util.Vector;
import java.lang.reflect.Constructor;

import com.yahoo.lsg.lifelong.GameInformation;
import com.yahoo.lsg.lifelong.LifelongStrategy;

import util.Array;
import lemonade.GeneralizedLemonadeUtility;
import game.Utility;

/**
 * Calculate the level-based strategies of the initial and state-based actions.
 * 
 *    This agent works by calculating the values resulting from the presence of 
 *    agents with random, best-response, next-level best response, and third-level 
 *    best-response (L0, L1, L2, and L3) strategies for the current game.
 *    Based on the observations over time, we classify opposing strategies 
 *    as some probability for each of the four levels.  
 *    We keep track of this distribution as totalLevelProbabilities.  
 *    In addition, we do the same for the state-based actions of a moving agent in 
 *    a given configuration, to estimate where they are likely to go once 
 *    the actions are played.  These levels are roughly classified as 
 *    random (L0), best response (L1), coalition strategy (L2), and coalition 
 *    strategy that maximizes score once the sucker responds (L3). 
 *    
 *    This technique outputs three candidate starting locations, which we 
 *    then analyze using the learned weights of the three levels in the population, 
 *    both initial and state-based strategies. 
 *    We run through the various starting configurations, weighted by probability, 
 *    and estimate the likely trajectory of the game over 100 rounds. 
 *    This method allows us to estimate the values of the three candidate locations. 
 *    As the initial action is the most important, the decision for the starting location 
 *    needs to utilize all available knowledge of the population, even though 
 *    the agent probably hasn't seen the particular payoff structure before. 
 *    Estimating the long-run values is important because we need to select an 
 *    action to provide the best chance for a high score over the course of 
 *    100 rounds, and not just the first configuration. 
 *    The level-based strategy is a good balance of generalizability and 
 *    customization for a specific payoff setting. 
 *    
 *    This agent does not attempt to identify the current opposing agents. 
 *    The reasoning is that either the opponents will be misclassified, 
 *    or the identification will come too late to make much difference. 
 *    Either way, we decided to make our decisions in response to the anonymous 
 *    population as a whole.  
 *    
 *    We also do not specifically test for learning opponents, 
 *    and so the agent may be executing its strategy based on old data. 
 *    An agent that successfully models its opponents will 
 *    gradually adapt its responses over time, and a population dominated by these 
 *    learning agents will take the form of a dynamical system which is
 *    difficult to estimate in the time frame provided.  
 *    However, the assumption is that either others are not learning as effectively, 
 *    or the population as a whole is not learning very quickly, 
 *    while still broadly playing the same basic strategies. 
 *    Under these conditions, our agent should build an accurate model.*
 *     
 *    (*Unless a better framework exists for these types of games.) 
 *    
 *     See Wunder et al. AAMAS 2012 for further details.
 *  
 *  
 * @author wunder
 */
public class AmazingMystery implements LifelongStrategy{
	  public final static int DEBUG = 0;
	  public final static int LOSS_THRESHOLD = 200;
	  public final static int SINGLE_TURN_THRESHOLD = 200;
	  
	  /**
	   * The action the player plays
	   */
  int constantAction;
  int turn = 0;
  int matchNumber = 0;

  int[] bestActions;
  int[] nextBestActions;
  
  int[] prevActions = new int[3];
  int[] originalActions = new int[3];
  int[][] initialCoalitions;
  int[] availableCoalitionActions;
  int[] densities;
  int totalUtility = 0;
  int numActions;
  private int mLastMovedTurn;	
  int[] scores = new int[3];
  int[] totalScores = new int[3];
  double[] stableScores = new double[3];
  
  boolean firstAction = true;
  boolean unchanged = true; 
  boolean original = true; 
  boolean intialToState = false;
  
  private Constructor<Strategy> mConstructor;
  private Strategy mStrategyInstance;
	
  public double[] L0strategy = new double[12];
  public double[] L1strategy = new double[12];
  public double[] L2strategy = new double[12];
  public double[] L3strategy = new double[12];
  public double[] L0stateStrategy = new double[12];
  public double[] L1stateStrategy = new double[12];
  public double[] L2stateStrategy = new double[12];
  public double[] L3stateStrategy = new double[12];
  public double[] totalLevelProbabilities = {0.0, 0.34, 0.33, 0.33};
  public double[] stateLevelProbabilities = new double[] {0.1, 0.4, 0.4, 0.1};
  public double[][] initialToStateLevel = new double[][] {{0.5, 0.5, 0.0, 0.0},{0.1, 0.4, 0.4, 0.1},{0.1, 0.4, 0.4, 0.1},{0.1, 0.4, 0.4, 0.1}};
  
  // Initial action levels: L0: random. L1: BR(L0,L0). L2: BR(L0,L1). L3: Max(L0,L1,BR(L0,L1))
  public static double[] levelProbabilities = new double[] {0.0, 0.25, 0.5, 0.25};
  // Action levels in state: L0: random. L1: BR(current). L2: look for coalition. L3: best coalition after sucker responds.
  public static double[][] tempStateLevelProbabilities = new double[2][5];// {{0.0, 0.25, 0.5, 0.25},{0.0, 0.25, 0.5, 0.25}};
  // Observed Initial Action Probabilities
  public static double[][] tempInitialLevelProbabilities = new double[2][5];// {{0.0, 0.25, 0.5, 0.25},{0.0, 0.25, 0.5, 0.25}};
  
  public double cumStayGains = 1;
  public double cumStayRegret = 1;
  public double expectedStayGains = 1;
  public double cumMoveGains = 1;
  public double cumMoveLoss = 1;
  public double expectedMoveLoss = 1;
  public double observedMoveStayProb = 0.75;
  public double observedStayStayProb = 0.5;
  public double learnedMoveStayProb = 0.75;
  public double learnedStayStayProb = 0.5;
  public int initialScore = 0;
  public int altStayScore = 0;
  public int projectedMoveScore = 0;
  public int altMoveScore = 0;
  public int switchGoodStayBad = 0;
  public int switchBadStayGood = 0;
  public boolean staySwitch = true;
  public int meBest = 0;
  
  private int[][][] mPrecalculatedUtility;
  
  GeneralizedLemonadeUtility utility;

  public double oGamma = 0.75; 
  public double moveGamma = 0.5; 
  public double gamma = 0.92; 
  public int cumRegret = 0; 
  public double temperature = 1.0;
  public double suckerGamma = 0.75; 
  public double learnRate = 0.01;
  
  public int suckerRegret = 0; 
  public int suckerStayGains = 0; 
  
  public int cumCoalitionRegret = 0; 
  public int projectedReward = 0; 
  
  // when deciding on our initial action, focus on these three strategies. 
  public int[] stableTriple = new int[3];
  public int[] stableTripleValues = new int[3];
  public int[] stableActionsFrequency = new int[3];
  public double[] estimatedValues = new double[3];
  public double valueError = 0.0; 
  int thresholdShift = 0;
  public Random r;
  
  public void newGame(Game g, int iterations){
	    this.numActions = g.getNumActions();
	    this.mLastMovedTurn = 1;
	    this.matchNumber++;
	    this.turn = 0;
	    this.firstAction = true;
	    this.utility = (GeneralizedLemonadeUtility)g.getUtility();
	    this.densities = utility.getIslandDensity();
	    initBestActions();
	    constantAction = nextBestActions[r.nextInt(nextBestActions.length)];
	    int[] tempActionArray = new int[3];
		int[] tempUtilityArray = new int[3];
		mPrecalculatedUtility = new int[numActions][numActions][numActions];
		this.suckerRegret = 0;
		this.cumRegret = 0;
		this.suckerStayGains = 0;
		this.cumCoalitionRegret = 0; 
		
		this.availableCoalitionActions = new int[numActions];
		for(int i=0; i<numActions; i++) {
			for(int j=i; j<numActions; j++) {
				for(int k=j; k<numActions; k++) {
					tempActionArray[0] = i;
					tempActionArray[1] = j;
					tempActionArray[2] = k;
					utility.getUtility(tempActionArray, tempUtilityArray);
					mPrecalculatedUtility[i][j][k] = tempUtilityArray[0];
					mPrecalculatedUtility[i][k][j] = tempUtilityArray[0];
					mPrecalculatedUtility[j][i][k] = tempUtilityArray[1];
					mPrecalculatedUtility[j][k][i] = tempUtilityArray[1];
					mPrecalculatedUtility[k][i][j] = tempUtilityArray[2];
					mPrecalculatedUtility[k][j][i] = tempUtilityArray[2];
				}
			}
		}   
		for (int i = 0; i<2; i++) {
			 this.tempStateLevelProbabilities[i][0] = 0.0;
			 this.tempStateLevelProbabilities[i][1] = 0.34;
			 this.tempStateLevelProbabilities[i][2] = 0.33;
			 this.tempStateLevelProbabilities[i][3] = 0.33;
			 this.tempStateLevelProbabilities[i][4] = 0.0;
			 this.tempInitialLevelProbabilities[i][0] = 0.1;
			 this.tempInitialLevelProbabilities[i][1] = 0.4;
			 this.tempInitialLevelProbabilities[i][2] = 0.4;
			 this.tempInitialLevelProbabilities[i][3] = 0.1;
		}
		for (int i = 0; i<3; i++) {
			stableScores[i]=0;
			scores[i]=0;
			totalScores[i]=0;
		}
		double totalP = 0.0;
		for (int i =1; i<4; i++)
			totalP += totalLevelProbabilities[i];
		for (int i =1; i<4; i++)
			levelProbabilities[i] = totalLevelProbabilities[i]/totalP;
		constantAction = returnActionAnalysis(bestActions[bestActions.length-1],nextBestActions[r.nextInt(nextBestActions.length)],oGamma);
		estimatedValues = projectedValue(stableTriple, new double[]{suckerGamma,suckerGamma});
		constantAction = stableTriple[returnMax(estimatedValues)];
	  }
  
  /**
   * Construct a strategy that builds a model of its opponents
   * and simulates the likely course of action, in order 
   * to compute the expected optimal action. 
   * 
   * @param g
   * @param seed
   */
  public AmazingMystery(GameInformation info, long seed, String options) {
	    this.r = new Random(seed);
	    for (int i = 0; i<L0strategy.length; i++) { 
			   L0strategy[i] = 1.0/L0strategy.length;
			   L0stateStrategy[i] = 1.0/L0strategy.length;
		}
  }  
	
  public int returnActionAnalysis(int action1, int action2, double otherGamma) {
	  stableTriple[0] = action1;
	  stableTriple[1] = action2;
	  stableTriple[2] = bestResponse(stableTriple[0], stableTriple[1]);
	  int bestAction = stableTriple[2];
	  stableTripleValues[0] = mPrecalculatedUtility[stableTriple[0]][stableTriple[1]][stableTriple[2]];
	  stableTripleValues[1] = mPrecalculatedUtility[stableTriple[1]][stableTriple[2]][stableTriple[0]];
	  stableTripleValues[2] = mPrecalculatedUtility[stableTriple[2]][stableTriple[0]][stableTriple[1]];
	  boolean useLearnedValues = false;
	  boolean rankLevels = false;  
	  double[] totalVals = new double[numActions];
	  
	  int maxTriple = 0;//stableTriple[0];
	  //identify the best action if these three actions are played
	  if (stableTripleValues[0]>stableTripleValues[1]) {
		  if (stableTripleValues[0]>stableTripleValues[2]) 
			  maxTriple = 0;//stableTriple[0];
		  else {
			  maxTriple = 2;//stableTriple[2];
		 }
	  }
	  else {
		  if (stableTripleValues[2]>stableTripleValues[1])
			  maxTriple = 2;//stableTriple[2];
		  else
			  maxTriple = 1;//stableTriple[1];
	  }
	  
	  int[] altRewards = new int[3];
	  double[] totalExpRewards = new double[3];
	  
	  altRewards[maxTriple] += mPrecalculatedUtility[stableTriple[maxTriple]][stableTriple[(maxTriple+1)%3]][stableTriple[maxTriple]];
	  altRewards[maxTriple] += mPrecalculatedUtility[stableTriple[maxTriple]][stableTriple[(maxTriple+2)%3]][stableTriple[maxTriple]];
	  altRewards[maxTriple] = altRewards[maxTriple]/2;
	  
	  altRewards[(maxTriple+1)%3] = mPrecalculatedUtility[stableTriple[(maxTriple+1)%3]][stableTriple[(maxTriple)%3]][stableTriple[maxTriple]];  
	  altRewards[(maxTriple+2)%3] = mPrecalculatedUtility[stableTriple[(maxTriple+2)%3]][stableTriple[(maxTriple)%3]][stableTriple[maxTriple]];  
	  
	  int potentialRegret = stableTripleValues[(maxTriple+1)%3]-mPrecalculatedUtility[stableTriple[maxTriple]][stableTriple[(maxTriple+2)%3]][stableTriple[maxTriple]];
	  potentialRegret = Math.max(potentialRegret,stableTripleValues[(maxTriple+2)%3]-mPrecalculatedUtility[stableTriple[maxTriple]][stableTriple[(maxTriple+1)%3]][stableTriple[maxTriple]]);
	  
	  int projectedR = stableTripleValues[(maxTriple)%3]-altRewards[maxTriple]; 
	   
	  double futureVal = (double)projectedR/(1.0-suckerGamma);
	  double t = futureVal/potentialRegret;
	  boolean includeProbs = true;
	  double[] locProbs = new double[3];
	  if (includeProbs) {
		  locProbs[0] = levelProbabilities[1];//1.0 - (1.0-levelProbabilities[1])*(1.0-levelProbabilities[1]);
		  locProbs[1] = levelProbabilities[2];//1.0 - (1.0-levelProbabilities[2])*(1.0-levelProbabilities[2]);
		  locProbs[2] = levelProbabilities[3];//1.0 - (1.0-levelProbabilities[3])*(1.0-levelProbabilities[3]);
		  totalExpRewards[maxTriple] = t*(locProbs[maxTriple]*altRewards[maxTriple]+(1-locProbs[maxTriple])*stableTripleValues[maxTriple]) + (100.0-t)*(double)stableTripleValues[maxTriple];
		  totalExpRewards[(maxTriple+1)%3] = t*(locProbs[(maxTriple+1)%3]*altRewards[(maxTriple+1)%3]+(1-locProbs[(maxTriple+1)%3])*stableTripleValues[(maxTriple+1)%3]) + (100.0-t)*(double)stableTripleValues[(maxTriple+1)%3];
		  totalExpRewards[(maxTriple+2)%3] = t*(locProbs[(maxTriple+2)%3]*altRewards[(maxTriple+2)%3]+(1-locProbs[(maxTriple+2)%3])*stableTripleValues[(maxTriple+2)%3]) + (100.0-t)*(double)stableTripleValues[(maxTriple+2)%3];
	  }
	  else {
		  totalExpRewards[maxTriple] = t*altRewards[maxTriple] + (100.0-t)*(double)stableTripleValues[maxTriple];
		  totalExpRewards[(maxTriple+1)%3] = t*altRewards[(maxTriple+1)%3] + (100.0-t)*(double)stableTripleValues[(maxTriple+1)%3];
		  totalExpRewards[(maxTriple+2)%3] = t*altRewards[(maxTriple+2)%3] + (100.0-t)*(double)stableTripleValues[(maxTriple+2)%3];
	  }
	  	
	  double[] strat1 = new double[numActions];
	  double[] strat2 = new double[numActions];
	  double[] strat3 = new double[numActions];
	  strat1[stableTriple[0]] = 1.0;
	  strat2[stableTriple[1]] = 1.0;
	  strat3[stableTriple[2]] = 1.0;
	  double tempTemp = 0.5; 
	  double[] eValues1 = expectedValues(strat3, strat2);
	  double[] eValues2 = expectedValues(strat1, strat3);
	  double[] eValues3 = expectedValues(strat2, strat1);
	  double expValue1 = eValues1[stableTriple[0]];
	  double expValue2 = eValues2[stableTriple[1]];
	  double expValue3 = eValues3[stableTriple[2]];
	  
	  if (rankLevels) {
		  eValues1 = expectedValues(L0strategy, L0strategy);
		  eValues2 = expectedValues(strat1, L0strategy);
		  eValues3 = expectedValues(strat2, strat1);

	  if (expValue1>=expValue2) {
		  if (expValue1>=expValue3) {
			  quantalResponse(L1strategy,eValues1,tempTemp);
			  if (expValue2>=expValue3) {
				  quantalResponse(L2strategy,eValues2,tempTemp);
				  quantalResponse(L3strategy,eValues3,tempTemp);
			  }
			  else {
				  quantalResponse(L2strategy,eValues3,tempTemp);
				  quantalResponse(L3strategy,eValues2,tempTemp);
			  }
		  }
		  else {
			  quantalResponse(L1strategy,eValues3,tempTemp);
			  quantalResponse(L2strategy,eValues1,tempTemp);
			  quantalResponse(L3strategy,eValues2,tempTemp);
		  }
	  }
	  else {
		  if (expValue2>=expValue3) {
			  quantalResponse(L1strategy,eValues2,tempTemp);
			  if (expValue1>=expValue3) {
				  quantalResponse(L2strategy,eValues1,tempTemp);
				  quantalResponse(L3strategy,eValues3,tempTemp);
			  }
			  else {
				  quantalResponse(L3strategy,eValues1,tempTemp);
				  quantalResponse(L2strategy,eValues3,tempTemp);
			  }
		  }
		  else {
			  quantalResponse(L1strategy,eValues3,tempTemp);
			  quantalResponse(L2strategy,eValues2,tempTemp);
			  quantalResponse(L3strategy,eValues1,tempTemp);
		  }
	  }
	  }
	  else {
		    quantalResponse(L1strategy,expectedValues(L0strategy,L0strategy),temperature);
		    quantalResponse(L2strategy,expectedValues(L0strategy,L1strategy),temperature);
		    if (expValue1>=expValue2) {
				  if (expValue1>=expValue3) 
					  quantalResponse(L3strategy,eValues1,temperature);
				  else 
					  quantalResponse(L3strategy,eValues3,temperature);
		    }
		    else {
		    	if (expValue2>=expValue3) 
		    		quantalResponse(L3strategy,eValues2,temperature);
				else 
					quantalResponse(L3strategy,eValues3,temperature);
		    }
	  }
	  learnedExpectedValues(totalVals);
	  if (useLearnedValues){
		    if (totalVals[stableTriple[0]]>totalVals[stableTriple[1]]) {
		    	if (totalVals[stableTriple[0]]>totalVals[stableTriple[2]]) {
		    		bestAction = stableTriple[0];
		    	}
		    	else {
		    		bestAction = stableTriple[2];
		    	}
		    }
		    else {
		    	if (totalVals[stableTriple[1]]>totalVals[stableTriple[2]]) {
		    		bestAction = stableTriple[1];
		    	}
		    	else 
		    		bestAction = stableTriple[2];
		    }
		    //bestAction = bestAct(totalVals);	
	  }
	  else {
	  if (totalExpRewards[maxTriple]>=totalExpRewards[(maxTriple+1)%3]) {
		  if (totalExpRewards[maxTriple]>=totalExpRewards[(maxTriple+2)%3]) 
			  bestAction = stableTriple[maxTriple];
		  else 
			  bestAction = stableTriple[(maxTriple+2)%3];
	  }
	  else {
		  if (totalExpRewards[(maxTriple+1)%3]>=totalExpRewards[(maxTriple+2)%3]) 
			  bestAction = stableTriple[(maxTriple+1)%3];
		  else 
			  bestAction = stableTriple[(maxTriple+2)%3];
	  }}
	  
	  return bestAction;
  }
  
  /** 
   * @return an estimate of the expected value for the incoming game, for each considered location. 
   * 
   */
  public double[] projectedValue(int[] locations, double[] discount) {
	  double[] projectedValues = new double[locations.length];
	  double duration = 0.0;
	  int brS = 0;
	  int bestValue = 0;
	  for (int loc = 0; loc<locations.length; loc++) { 
		  double tempValue = 0.0;
		  double longValue = 0.0;
		  for (int i = 1; i<4; i++) {
			  for (int j = 1; j<4; j++) {
				  tempValue = mPrecalculatedUtility[locations[loc]][locations[i-1]][locations[j-1]];
				
				  if (loc==i-1) {
					  brS = bestResponse(locations[loc],locations[j-1]);
					  longValue = longTermValue(new int[]{locations[loc],locations[i-1],locations[j-1]},1,i);//					  longValue = mPrecalculatedUtility[locations[loc]][brS][locations[j-1]];
					  bestValue = mPrecalculatedUtility[brS][locations[loc]][locations[j-1]];
					  if ((bestValue-tempValue)*(1-discount[0])<=0) 
						  duration = 1;
					  else
						  duration = (double)(longValue-tempValue)/((bestValue-tempValue)*(1-discount[0]));
				  }
				  if (loc==j-1)  {
					  brS = bestResponse(locations[loc],locations[i-1]);
					  longValue = longTermValue(new int[]{locations[loc],locations[i-1],locations[j-1]},2,j);//					  longValue = mPrecalculatedUtility[locations[loc]][brS][locations[i-1]];
					  bestValue = mPrecalculatedUtility[brS][locations[loc]][locations[i-1]];
					  if ((bestValue-tempValue)*(1-discount[0])<=0) 
						  duration = 1;
					  else
						  duration = (double)(longValue-tempValue)/((bestValue-tempValue)*(1-discount[1]));
				  }
				  if (i==j) {
					  brS = bestResponse(locations[loc],locations[j-1]);
					  longValue = longTermValue(new int[]{locations[loc],locations[i-1],locations[j-1]},1,i);//					  longValue = mPrecalculatedUtility[locations[loc]][brS][locations[j-1]];
					  bestValue = mPrecalculatedUtility[brS][locations[loc]][locations[j-1]];
					  if ((bestValue-tempValue)*(1-discount[0])<=0) 
						  duration = 1;
					  else
						  duration = (double)(mPrecalculatedUtility[locations[j-1]][brS][locations[loc]]-mPrecalculatedUtility[locations[i-1]][locations[j-1]][locations[loc]])/((bestValue-mPrecalculatedUtility[locations[i-1]][locations[j-1]][locations[loc]])*(1-discount[0]));
				  }
				  projectedValues[loc] += levelProbabilities[i]*levelProbabilities[j]*(tempValue*duration+longValue*(100.0-duration));
			  }  
		  }
		  projectedValues[loc] = projectedValues[loc]/(1.0-levelProbabilities[0]);
	  }
	  return projectedValues;
  }
  
  /** 
   * @return an estimate of the expected value of an action in the current state 
   * 
   */
  public double[] projectedStateValue(int[] locations, double[] discount) {
	  double[] projectedValue = new double[numActions];
	  double bestValue = 0;
	  double tempValue = 0.0;
	  double longValue = 0.0;
	  double duration = 0.0;
	  int brS = -1;
	  
	  for (int loc = 0; loc<locations.length; loc++) { 
		  tempValue = mPrecalculatedUtility[loc][locations[1]][locations[2]];
		  
		  if (mPrecalculatedUtility[locations[1]][loc][locations[2]]<mPrecalculatedUtility[locations[2]][loc][locations[1]]) {
			  brS = bestResponse(loc,locations[2]);
			  bestValue = mPrecalculatedUtility[brS][loc][locations[2]];
			  for (int level = 1; level<4; level++) {
				  longValue = longTermValue(new int[]{loc,locations[1],locations[2]}, 1, level);
				  if ((bestValue-tempValue)*(1-discount[0])<=0) 
					  duration = 1;
				  else
					  duration =  (longValue-tempValue)/((bestValue-tempValue)*(1.0-discount[0]));
			 	  projectedValue[loc] += tempInitialLevelProbabilities[0][level]*(tempValue*duration+longValue*(turn-duration));
			  }
		  }
		  else {
			  brS = bestResponse(loc,locations[1]);
			  bestValue = mPrecalculatedUtility[brS][loc][locations[1]];
			  for (int level = 1; level<4; level++) {
				  longValue = longTermValue(new int[]{loc,locations[1],locations[2]}, 2, level);
				  if ((bestValue-tempValue)*(1-discount[0])<=0) 
					  duration = 1;
				  else
					  duration =  (longValue-tempValue)/((bestValue-tempValue)*(1.0-discount[1]));
			 	  projectedValue[loc] += tempInitialLevelProbabilities[1][level]*(tempValue*duration+longValue*(turn-duration));
			  }
		  }
		  projectedValue[loc] = projectedValue[loc]/(100-turn);
	  }
	  return projectedValue; 
  }
  
  /** 
   * 
   * @param locations: the three locations of the agents
   * @param mover: which agent is the one we expect would move to a different location
   * @return  Calculate the long-term value of a given configuration
   */
  public double longTermValue(int[] locations, int mover, int level) {
	  double longValue = 0.0;
	  int brS = 0; 
	  int coalitionAct = 0;
	  int coalitionActBR = 0;
	  double[] s1 = new double[numActions];
	  double[] s2 = new double[numActions];
	  double[] evaluesBR =new double[numActions];
	  double[] evaluesCoalition =new double[numActions];
	  double[] evaluesCoalitionBR =new double[numActions];
	  int[] testCoalitionActions = new int[numActions];
	  int moverAction = 0;
	  
	  s1[locations[0]] = 1.0;
	  s2[locations[3-mover]] = 1.0;
	  brS = bestResponse(locations[0],locations[3-mover]);
	  evaluesBR = expectedValues(s1,s2);
	  
	  currentCoalitions(new int[]{locations[mover],locations[0],locations[3-mover]},testCoalitionActions);
	  for (int i = 0; i<evaluesCoalition.length; i++) {
		  if (testCoalitionActions[i]>0) {
			  evaluesCoalition[i]=evaluesBR[i];
			  if (mPrecalculatedUtility[prevActions[0]][prevActions[3-mover]][i]>mPrecalculatedUtility[prevActions[3-mover]][prevActions[0]][i]) {
				  moverAction = bestResponse(prevActions[0],i);
				  evaluesCoalitionBR[i]=mPrecalculatedUtility[i][prevActions[0]][moverAction];
			  }
			  else {
				  moverAction = bestResponse(prevActions[3-mover],i);
				  evaluesCoalitionBR[i]=mPrecalculatedUtility[i][prevActions[3-mover]][moverAction];
			  }
		  }
	  }
	  coalitionActBR = returnMax(evaluesCoalitionBR);
	  coalitionAct = returnMax(evaluesCoalition);
	  
	  if (intialToState) {
		  longValue = initialToStateLevel[level][1]*mPrecalculatedUtility[locations[0]][locations[3-mover]][brS]+
				  	  initialToStateLevel[level][2]*mPrecalculatedUtility[locations[0]][locations[3-mover]][coalitionAct]+
				  	  initialToStateLevel[level][3]*mPrecalculatedUtility[locations[0]][locations[3-mover]][coalitionActBR];
		  longValue = longValue/(1-initialToStateLevel[level][0]);
	  }
	  else {
		  longValue = stateLevelProbabilities[1]*mPrecalculatedUtility[locations[0]][locations[3-mover]][brS]+
				  	  stateLevelProbabilities[2]*mPrecalculatedUtility[locations[0]][locations[3-mover]][coalitionAct]+
				  	  stateLevelProbabilities[3]*mPrecalculatedUtility[locations[0]][locations[3-mover]][coalitionActBR];
		  longValue = longValue/(1-stateLevelProbabilities[0]);
	  }
	  return longValue;
  }
  
  public int bestAct(double[] payoffs) {
	  int bestA = 0;
	  double bestV = 0.0;
	  for (int i = 0; i<numActions; i++) {
		  if (payoffs[i]>bestV) {
			  bestV = payoffs[i];
			  bestA = i;
		  }
	  }
	  return bestA; 
  }
  
  public double[] quantalResponse(double[] strategy, double[] values, double temp) {
	  double total = 0.0;
	  for (int i=0; i<values.length; i++)
		  total+=Math.exp(temp*values[i]);
	  for (int i=0; i<values.length; i++)
		  strategy[i] = Math.exp(temp*values[i])/total;
	  return strategy;
  }
  
  public double[] expectedValues(double[] strategy1, double[] strategy2) {
	  double[] eValues = new double[strategy1.length];
	  for (int i=0; i<strategy1.length; i++) {
		  for (int j=0; j<strategy1.length; j++) {
				for (int k=0; k<strategy1.length; k++) {
						int[] util = utility.getUtility(new int[]{i,j,k});
						eValues[i] += strategy1[j]*strategy2[k]*util[0];
				}
		  }
	  }
	  return eValues;
  }

  public int returnMax(double[] vals) {
	  double maxVal = 0.0;
	  int maxIndex = 0;
	  for (int i = 0; i<vals.length; i++) {
		  if (vals[i]>maxVal) {
			  maxVal = vals[i];
			  maxIndex = i;
		  }
	  }
	  return maxIndex;
  }
  
  // Use Bayes here to determine the distribution over levels. 
  public double[] calculateProbabilities(int act, double[] strategyDist) {
	strategyDist[0] = L0strategy[act]/(L0strategy[act]+L1strategy[act]+L2strategy[act]+L3strategy[act]);
	strategyDist[1] = L1strategy[act]/(L0strategy[act]+L1strategy[act]+L2strategy[act]+L3strategy[act]);
	strategyDist[2] = L2strategy[act]/(L0strategy[act]+L1strategy[act]+L2strategy[act]+L3strategy[act]);
	strategyDist[3] = L3strategy[act]/(L0strategy[act]+L1strategy[act]+L2strategy[act]+L3strategy[act]);
  	return strategyDist;
  }
  
  public double[] calculateStateProbabilities(int act, double[] strategyDist) {
	strategyDist[0] = L0stateStrategy[act]/(L0stateStrategy[act]+L1stateStrategy[act]+L2stateStrategy[act]+L3stateStrategy[act]);
	strategyDist[1] = L1stateStrategy[act]/(L0stateStrategy[act]+L1stateStrategy[act]+L2stateStrategy[act]+L3stateStrategy[act]);
	strategyDist[2] = L2stateStrategy[act]/(L0stateStrategy[act]+L1stateStrategy[act]+L2stateStrategy[act]+L3stateStrategy[act]);
	strategyDist[3] = L3stateStrategy[act]/(L0stateStrategy[act]+L1stateStrategy[act]+L2stateStrategy[act]+L3stateStrategy[act]);
  	return strategyDist;
  }
 
  public double[][] updateInitialToStateMapping() {
	  for (int p = 0; p<2; p++) {
		  for (int i = 0; i<4; i++) {
			  for (int j = 0; j<4; j++) {
				  initialToStateLevel[i][j] = //(1.0-learnRate*tempInitialLevelProbabilities[p][i])*initialToStateLevel[i][j] + 
						  						//   learnRate*tempInitialLevelProbabilities[p][i]*tempStateLevelProbabilities[p][j];
						  ((Math.min(100,matchNumber)+1.0-tempInitialLevelProbabilities[p][i])*initialToStateLevel[i][j] + 
	  						   tempInitialLevelProbabilities[p][i]*tempStateLevelProbabilities[p][j])/(Math.min(100,matchNumber)+1.0);
			  }
		  }
	  }  
	  return initialToStateLevel;
  }
  
  public int bestResponse(int act1, int act2) {
	  int maxVal = 0;
	  int maxAct = 0;
	  for (int i = 0; i<numActions; i++) {
		  if (mPrecalculatedUtility[i][act1][act2] > maxVal) {
			  maxAct = i;
			  maxVal = mPrecalculatedUtility[i][act1][act2];
		  }
	  }
	  return maxAct;
  }
  
  public double[] learnedExpectedValues(double[] totalVals) {
	  double[][] valueHolder = new double[4][numActions]; 
	  
	  System.arraycopy(expectedValues(L0strategy,L0strategy),0,valueHolder[0],0,numActions);
	  System.arraycopy(expectedValues(L0strategy,L1strategy),0,valueHolder[1],0,numActions);
	  System.arraycopy(expectedValues(L0strategy,L2strategy),0,valueHolder[2],0,numActions);
	  System.arraycopy(expectedValues(L0strategy,L3strategy),0,valueHolder[3],0,numActions);
	  
	  for (int level = 0; level<4; level++) {
		  for (int i = 0; i<numActions; i++) {
				  totalVals[i]+=levelProbabilities[level]*valueHolder[level][i];
		  }
	  }
	  
	  return totalVals;
  }
  
  protected int getQuality(int action){
	    int totalValue = 0;
	    int oppositeAction = (action+utility.getIslandSize()/2)% utility.getIslandSize();
	    int[] util = utility.getUtility(new int[]{action,oppositeAction,oppositeAction});
	    for (int i = 0; i<utility.getIslandSize(); i++) {
		  for (int j = 0; j<utility.getIslandSize(); j++) {
			util = utility.getUtility(new int[]{action,i,j});
			totalValue += util[0];
		  }
	    }
	    return totalValue;
	  }

  protected int getNextBestQuality(int action, int otherAction) {
		int totalValue = 0;
	    int[] util;
	    for (int i = 0; i<utility.getIslandSize(); i++) {
			util = utility.getUtility(new int[]{action,i,otherAction});
			totalValue += util[0];
	    }
	    return totalValue;
	  }
	  
  private int[] initBestActions() {
	    int[] quality = new int[utility.getIslandSize()];
	    int[] qualityNB = new int[utility.getIslandSize()];
	    for(int i=0;i<quality.length;i++){
	      quality[i]=getQuality(i);
	    }
	    int maxDensity = quality[0];
	    int maxDensityCount = 1;
	    for(int i=0;i<quality.length;i++){
	      if (maxDensity<quality[i]){
	        maxDensity=quality[i];
	        maxDensityCount=1;
	      } else if (maxDensity==quality[i]){
	        maxDensityCount++;
	      }
	    }
	    bestActions = new int[maxDensityCount];
	    int bestActionIndex = 0;
	    for(int i=0;i<quality.length;i++){
	      if (quality[i]==maxDensity){
	        bestActions[bestActionIndex]=i;
	        bestActionIndex++;
	      }
	    }

	    for (int b = 0; b<bestActions.length; b++) {
	    	for(int i=0;i<qualityNB.length;i++){
	    		qualityNB[i]=getNextBestQuality(i,bestActions[b]);
	    	}
	    }
	    
	    
	    maxDensity = qualityNB[0];
	    maxDensityCount = 1;
	    for(int i=0;i<qualityNB.length;i++){
	      if (maxDensity<qualityNB[i]){
	        maxDensity=qualityNB[i];
	        maxDensityCount=1;
	      } else if (maxDensity==qualityNB[i]){
	        maxDensityCount++;
	      }
	    }
	    nextBestActions = new int[maxDensityCount];
	    int nextBestActionIndex = 0;
	    for(int i=0;i<quality.length;i++){
	      if (qualityNB[i]==maxDensity){
	        nextBestActions[nextBestActionIndex]=i;
	        nextBestActionIndex++;
	      }
	    }
		
	    return quality; 
	  }
	  
  protected int qualityRange(int[] qual) {
		  int total = 0;
		  int max = 0;
		  for (int i=0; i<qual.length; i++) {
			  total+=qual[i];
			  if (max>qual[i]) 
				  max = qual[i];
		  }
		  return max - total/numActions;		  
	  }
	  
  // Are X and Y in a "coalition" against Z? 
  // Do they get higher scores than Z and 
  // relatively equal to each other? 
  protected boolean coalitionTest(int[] util) {
	return (!(util[2] > util[0]));
	//return (!((util[2] > util[0]) || (util[2] > util[1])));
  }
  
  //return an array of the possible coalition-building spots. 
  //0 = none, 1 = coalition with player 1, 2 = coalition w/P2, 3 = BOTH
  public int[] currentCoalitions(int[] actions, int[] coalitions) {
	boolean coalitionI = true; 
	boolean coalitionJ = true;
	int[] origUtil = utility.getUtility(new int[]{actions[0],actions[1],actions[2]});
	for (int i = 0; i< numActions; i++) {
	   coalitionI = true;
  	   coalitionJ = true;
	   
 	   int[] util = utility.getUtility(new int[]{i,actions[1],actions[2]});
	   int[] util1;
	   int[] util2;
		int maxScoreI = util[2];
		int maxScoreJ = util[1];

		for (int k = 0; k<numActions; k++) {
			util1 = utility.getUtility(new int[]{i,actions[1],k});
			util2 = utility.getUtility(new int[]{i,actions[2],k});
			if (util1[2] > maxScoreI) 
				maxScoreI = util1[2];
			if (util2[1] > maxScoreJ) 
				maxScoreJ = util2[1]; 
		}
		for (int k = 0; k<numActions; k++) {
			util1 = utility.getUtility(new int[]{i,actions[1],k});
			util2 = utility.getUtility(new int[]{i,actions[2],k});
			coalitionI = (coalitionI&&coalitionTest(util1));//new int[]{i,actions[1],k}));
			coalitionJ = (coalitionJ&&coalitionTest(util2));//new int[]{i,actions[2],k}));
		}
	   
	   if (coalitionI) 
		   coalitions[i]++;
	   if (coalitionJ)
		   coalitions[i]+=2;
	   }
	return coalitions;
  }

  public int getAction() {
	boolean useEstimatedValues = false; 
    int[] util = utility.getUtility(prevActions);
    int bestValue = util[0];
    int bestAction = constantAction; 
    int[] utilC = new int[3];
    int[] utilAlt = new int[3];
    int[] utilAltMoved = new int[3];
    
    for (int i = 0; i< 3; i++) {
		scores[i]+=util[i];
		totalScores[i]+=util[i];
		stableScores[i]+= mPrecalculatedUtility[stableTriple[i]][prevActions[1]][prevActions[2]];
    }
    
    if (scores[0]>scores[1] && scores[0]>scores[2]) {
    	if (unchanged)
    		return constantAction;
    	else {
    		for (int i = 0; i< 3; i++) 
    			scores[i] = 0;
    		cumRegret = 0;
    		cumCoalitionRegret = 0;
    	}
    }
    
    if (!unchanged) {
    	for (int i = 0; i<availableCoalitionActions.length; i++) {
    		availableCoalitionActions[i] = 0;
    	}
            currentCoalitions(prevActions,availableCoalitionActions);
    }
    for (int i = 0; i<availableCoalitionActions.length; i++) {
    	if (availableCoalitionActions[i]>0) {
    		utilC = utility.getUtility(new int[]{i,prevActions[1],prevActions[2]});
    		if (utilC[0]>bestValue) {
    			bestValue = utilC[0];
    			bestAction = i;
    		}
    	}
    }
    
    int sucker = 0;
    if (util[1]<=util[2])
    	sucker = 1;
    else 
    	sucker = 2;
    int otherBest = 0;
    int otherValue = 0;
    for (int i = 0; i<numActions; i++) {
    		utilAlt = utility.getUtility(new int[]{i,prevActions[3-sucker],prevActions[0]});
    		if (utilAlt[0]>otherValue) {
    			otherValue = utilAlt[0];
    			otherBest = i;
    		}
    }
    utilAlt = utility.getUtility(new int[]{prevActions[0],prevActions[3-sucker],otherBest});
    
    int bestA = bestPayoffs(currentPayoffs(prevActions));
	int[] bestU = utility.getUtility(new int[]{bestAction,prevActions[1],prevActions[2]});
	
	if (bestU[1]<=bestU[2])
    	sucker = 1;
    else 
    	sucker = 2;
    otherValue = 0;
    for (int i = 0; i<numActions; i++) {
    	utilAltMoved = utility.getUtility(new int[]{i,prevActions[3-sucker],bestAction});
    		if (utilAltMoved[0]>otherValue) {
    			otherValue = utilAltMoved[0];
    			otherBest = i;
    		}
    }
        
    utilAltMoved= utility.getUtility(new int[]{bestAction,prevActions[3-sucker],otherBest});
    cumCoalitionRegret += bestValue - util[0];
    cumRegret += bestU[0] - util[0];
    suckerRegret += mPrecalculatedUtility[bestResponse(prevActions[0],prevActions[3-sucker])][prevActions[0]][prevActions[3-sucker]]-util[sucker];
    suckerStayGains = mPrecalculatedUtility[prevActions[sucker]][bestResponse(prevActions[1],prevActions[2])][prevActions[3-sucker]]-util[sucker];
    
    projectedReward = utilAlt[0] - util[0];
    if (projectedReward<0)
    	projectedReward = 0;
    if (cumCoalitionRegret > (projectedReward)/(1-gamma) &&  bestValue > util[0] && turn > 1) {
    //if (cumCoalitionRegret > (util[0]+projectedReward)/(1-gamma) &&  bestValue > util[0] && turn > 1) {
    //System.out.println(Math.exp(Math.log(Math.max(0.01,(bestValue - util[0])/Math.max(1,bestValue - util[0]+utilAlt[0]-utilAltMoved[0])))/turn));
	//if (gamma < Math.exp(Math.log(Math.max(0.01,(bestValue - util[0])/Math.max(1,bestValue - util[0]+utilAlt[0]-utilAltMoved[0])))/(turn-mLastMovedTurn))  &&  bestValue > util[0] && turn > 1) {
    //if ((util[0]/(1-observedStayStayProb)+utilAlt[0]*(100-1.0/(1-observedStayStayProb)) <
     // (bestValue/(1-observedMoveStayProb)+utilAltMoved[0]*(100-1.0/(1-observedMoveStayProb)))) &&  bestValue > util[0] && turn > 1) {
    		
    	double[] longV;
		if (useEstimatedValues) {
    		longV = projectedStateValue(prevActions,new double[]{suckerGamma,suckerGamma});
    		constantAction = returnMax(longV);
    		if (longV[constantAction]<bestValue)
    			constantAction = bestAction;
		}
    	else
    		constantAction = bestAction; 
	    for (int i = 0; i< 3; i++) 
			scores[i]=0;
		cumRegret = 0;
		cumCoalitionRegret = 0;
    }
    else {
    	if (bestU[1]<=bestU[2])
        	sucker = 1;
        else 
        	sucker = 2;
        otherValue = 0;
        for (int i = 0; i<numActions; i++) {
        	utilAltMoved = utility.getUtility(new int[]{i,prevActions[3-sucker],bestAction});
        		if (utilAltMoved[0]>otherValue) {
        			otherValue = utilAltMoved[0];
        			otherBest = i;
        		}
        }
        utilAltMoved= utility.getUtility(new int[]{bestAction,prevActions[3-sucker],otherBest});
        			    
        if (cumRegret > (projectedReward)/(1-gamma) &&  bestU[0] > util[0] && turn > 1) {
        //if (cumRegret > (util[0]+projectedReward)/(1-gamma) &&  bestU[0] > util[0] && turn > 1) {
    	//if (gamma < Math.exp(Math.log(Math.max(0.01,(bestU[0] - util[0])/Math.max(1,bestU[0] - util[0]+utilAlt[0]-utilAltMoved[0])))/(turn-mLastMovedTurn)) &&  bestU[0] > util[0] && turn > 1) {
        //if ((util[0]/(1-observedStayStayProb)+utilAlt[0]*(100-turn-1.0/(1-observedStayStayProb)) <
        //  (bestU[0]/(1-observedMoveStayProb)+utilAltMoved[0]*(100-turn-1.0/(1-observedMoveStayProb)))) &&  bestU[0] > util[0] && turn > 1) {
        	double[] longV;
        	if (useEstimatedValues) {
        		longV = projectedStateValue(prevActions,new double[]{suckerGamma,suckerGamma});
        		constantAction = returnMax(longV);
        		if (longV[constantAction]<bestU[0])
        			constantAction = bestA;
        	}
        	else
        		constantAction = bestA; 
	    for (int i = 0; i< 3; i++) 
			scores[i]=0;
		cumRegret = 0;
		cumCoalitionRegret = 0;
    	}
    }

	return constantAction;
	
  }

  // Action levels in state: L0: random. L1: BR(current). L2: look for coalition. L3: best coalition after sucker responds.
  public double[] analyzeStateLevels(int[] actions, double[][] stateLevels) {
	  double[] s1 = new double[numActions];
	  double[] s2 = new double[numActions];
	  double[] evaluesBR =new double[numActions];
	  double[] evaluesCoalition =new double[numActions];
	  double[] evaluesCoalitionBR =new double[numActions];
	  double[] tempLevels = new double[stateLevels[0].length-1];
	  int[] testCoalitionActions = new int[numActions];
	  int moverAction = 0;
	  int mover = 1;
	  if (actions[2]!=prevActions[2]) {
		  mover = 2;
	  }
	  s1[actions[0]]=1.0;
	  s2[actions[3-mover]]=1.0;
	   
	  evaluesBR = expectedValues(s1,s2);
	  
	  currentCoalitions(prevActions,testCoalitionActions);
	  for (int i = 0; i<evaluesCoalition.length; i++) {
		  if (testCoalitionActions[i]>0) {
			  evaluesCoalition[i]=evaluesBR[i];
			  if (mPrecalculatedUtility[prevActions[0]][prevActions[3-mover]][i]>mPrecalculatedUtility[prevActions[3-mover]][prevActions[0]][i]) {
				  moverAction = bestResponse(prevActions[0],i);
				  evaluesCoalitionBR[i]=mPrecalculatedUtility[i][prevActions[0]][moverAction];
			  }
			  else {
				  moverAction = bestResponse(prevActions[3-mover],i);
				  evaluesCoalitionBR[i]=mPrecalculatedUtility[i][prevActions[3-mover]][moverAction];
			  }
		  }
	  }
	  quantalResponse(L1stateStrategy,evaluesBR,1.0);
	  quantalResponse(L2stateStrategy,evaluesCoalition,1.0);
	  quantalResponse(L3stateStrategy,evaluesCoalitionBR,1.0);
	  calculateStateProbabilities( actions[mover], tempLevels);
	  for (int i =0; i<tempLevels.length; i++) 
		  stateLevels[mover-1][i] = (stateLevels[mover-1][4]*stateLevels[mover-1][i] + tempLevels[i])/(stateLevels[mover-1][4]+1);
	  stateLevels[mover-1][4]++;
	  return stateLevels[mover-1];
  }
  
  /*
   *  Set the tracking variable for cumulative score counts: 
   *  	public double cumStayGains;
  		public double cumStayRegret;
  		public double expectedStayGains;
  		public double cumMoveGains;
  		public double cumMoveLoss;
  		public double expectedMoveLoss;
   */
  public void adjustCumulativeTotals(int[] actions) {
	  int[] util = utility.getUtility(originalActions);
	  int br0 = bestResponse(actions[1],actions[2]);
	  int br1 = bestResponse(actions[0],actions[2]);
	  int br2 = bestResponse(actions[1],actions[0]);
	  if (turn==1) {
			initialScore = mPrecalculatedUtility[actions[0]][actions[1]][actions[2]];
			if (util[0]<=util[1] || util[0]<=util[2]) 
				expectedStayGains+=-util[0]+Math.min(mPrecalculatedUtility[actions[0]][br1][actions[2]],mPrecalculatedUtility[actions[0]][actions[1]][br2]);
	  }
	  else {
		  if (util[0]<=util[1] || util[0]<=util[2]) {
		    	if (firstAction) {
		    		if (original) {
		    			cumStayRegret+=-util[0]+mPrecalculatedUtility[br0][actions[1]][actions[2]];
		    			
		    		}
		    		else {
		    			cumStayGains+=util[0]-mPrecalculatedUtility[originalActions[0]][actions[1]][actions[2]];
		    		}
		    	}
		    	else {
		    		if (original) {
		    			cumMoveGains+=util[0]-mPrecalculatedUtility[actions[0]][actions[1]][actions[2]];
		    			if (!unchanged)
		    				expectedMoveLoss+= -util[0]+Math.min(mPrecalculatedUtility[actions[0]][br1][actions[2]],mPrecalculatedUtility[actions[0]][actions[1]][br2]);
		    		}
		    		else {
		    			cumMoveLoss+=util[0]-mPrecalculatedUtility[actions[0]][originalActions[1]][originalActions[2]];
		    		}
		    	}
		  }
	  }
	  
	  moveGamma = Math.max(0,Math.min(1,(cumStayGains-cumStayRegret-expectedStayGains)/Math.max(1,cumStayGains-cumStayRegret)));
	  moveGamma = Math.max(0,Math.min(1,(cumMoveGains-cumMoveLoss-expectedMoveLoss)/Math.max(1,cumMoveGains-cumMoveLoss)));
	  return;
  }
  
  public boolean stayOrSwitch(int[] actions, int[] utils) {
	  int br = bestResponse(actions[1],actions[2]);
	  int brS = 0;
	  int stayReward = 0;
	  
	  if (utils[1]>utils[2]) {
		  brS = bestResponse(actions[2],actions[0]);
		  stayReward = mPrecalculatedUtility[actions[0]][actions[1]][brS];
	  }
	  else {
		  brS = bestResponse(actions[1],actions[0]);
		  stayReward = mPrecalculatedUtility[actions[0]][brS][actions[2]];
	  }
	  return (mPrecalculatedUtility[br][actions[1]][actions[2]]<stayReward);
  }

  public void observeOutcome(int[] actions) {
	  double[] strategies0 = new double[4];
	  double[] totalVals = new double[numActions];
	  
	  if (DEBUG==-1) {
		  printGame(actions);
	  }
	  if (turn==1) {
		  if (mPrecalculatedUtility[actions[0]][actions[1]][actions[2]]>=mPrecalculatedUtility[actions[1]][actions[0]][actions[2]]&&mPrecalculatedUtility[actions[0]][actions[1]][actions[2]]>=mPrecalculatedUtility[actions[2]][actions[1]][actions[0]])
			  meBest++;
		  else {
			  if (stayOrSwitch(actions,new int[]{mPrecalculatedUtility[actions[0]][actions[1]][actions[2]],mPrecalculatedUtility[actions[1]][actions[0]][actions[2]],mPrecalculatedUtility[actions[2]][actions[1]][actions[0]]})) {
				  staySwitch = true;
				  switchBadStayGood++;
			  }
			  else {
				  switchGoodStayBad++;
				  staySwitch = false;
			  }
		  }
		  calculateProbabilities(actions[0], strategies0);
		  calculateProbabilities(actions[1], tempInitialLevelProbabilities[0]);
		  calculateProbabilities(actions[2], tempInitialLevelProbabilities[1]);
		  for (int i=0; i<4; i++) {
			  totalLevelProbabilities[i] += tempInitialLevelProbabilities[0][i];
			  totalLevelProbabilities[i] += tempInitialLevelProbabilities[1][i];
		  }
		  
		  if (matchNumber%100==11) {
			  learnedExpectedValues(totalVals);
		  }
		  for (int i=0; i<3; i++)
			  originalActions[i] = actions[i];
	  }
	  if (turn >= 99) {
		  int origAction = 0;
		  if (originalActions[0]==stableTriple[1])
			  origAction = 1;
		  if (originalActions[0]==stableTriple[2])
			  origAction = 2;
		  for (int i=0; i<4; i++) {
			  stateLevelProbabilities[i] = ((Math.min(1000,matchNumber))*stateLevelProbabilities[i] + 
					  						(tempStateLevelProbabilities[0][4]*tempStateLevelProbabilities[0][i]+ 
					  						 tempStateLevelProbabilities[1][4]*tempStateLevelProbabilities[1][i])/
					  						 (Math.max(1,tempStateLevelProbabilities[0][4]+tempStateLevelProbabilities[1][4])) )/
					  						 (Math.min(1000,matchNumber)+1.0);
		  }
		  updateInitialToStateMapping();
		  valueError += (-valueError + Math.pow(estimatedValues[origAction]/100.0-stableScores[origAction]/100.0,2.0))/matchNumber;
	  }
	  
	  int sucker = 0;
	    
	  if (mPrecalculatedUtility[prevActions[1]][prevActions[0]][prevActions[2]]<=mPrecalculatedUtility[prevActions[2]][prevActions[0]][prevActions[1]])
		  sucker = 1;
	  else 
		  sucker = 2;
	    
	  if ((actions[1]!=originalActions[1]||actions[2]!=originalActions[2]) && turn>0) {
		  if (original) {
			  if (firstAction) {
				  observedStayStayProb = ((matchNumber)*observedStayStayProb+((double)turn-1.0)/Math.max(1,turn))/(matchNumber+1.0);
				  if (suckerRegret > 0 && suckerRegret>suckerStayGains && suckerStayGains>=0) 
					  suckerGamma = ((Math.min(1000,matchNumber))*suckerGamma + (suckerRegret-suckerStayGains)/(suckerRegret))/(Math.min(1000,matchNumber)+1.0);
				  else {
					  if (stayOrSwitch(new int[] {actions[sucker],actions[0],actions[3-sucker]}, new int[] {mPrecalculatedUtility[actions[0]][actions[1]][actions[2]],mPrecalculatedUtility[actions[1]][actions[0]][actions[2]],mPrecalculatedUtility[actions[2]][actions[0]][actions[1]]}))
						  suckerGamma = (Math.min(1000,matchNumber)*suckerGamma)/(Math.min(1000,matchNumber)+1.0);
				  }
			  }
			  else  {
				  observedMoveStayProb = (Math.min(100,matchNumber)*observedMoveStayProb+((double)Math.max(1.0,turn-mLastMovedTurn)-1.0)/Math.max(1,turn-mLastMovedTurn))/Math.min(101.0,matchNumber+1.0);
			  }
		  }
		  mLastMovedTurn = turn;
		  if (unchanged) {
			  if (firstAction) {
				  learnedStayStayProb += (-learnedStayStayProb+observedStayStayProb)/(switchBadStayGood+1.0);
			  }
			  else {
				  learnedStayStayProb += (-learnedStayStayProb+((double)turn)/Math.max(1,turn+1.0))/(switchBadStayGood+1.0);
			  }
		  }
		  
	  }
	  original = (actions[1]==originalActions[1]&&actions[2]==originalActions[2]);
	  firstAction = firstAction&&(actions[0]==prevActions[0]);
	  unchanged = (actions[0]==prevActions[0]&&actions[1]==prevActions[1]&&actions[2]==prevActions[2]);
	  if (actions[1]!=prevActions[1]||actions[2]!=prevActions[2]) {
		  
		  analyzeStateLevels(actions,tempStateLevelProbabilities);
	  }
	  adjustCumulativeTotals(actions);
	  turn++;
	  for (int i=0; i<actions.length; i++) 
		  prevActions[i] = actions[i];
	  
  }
  
  // find the payoffs of a given P0/P1 arrangement, with P2 playing randomly
  protected int[] uniformPayoffs(int[] actions) {
	  int[] util;
	  int[] totalUtil = new int[numActions];
	  for (int i = 0; i<numActions; i++) {
		  for (int k = 0; k<numActions; k++) {
	  		  util= utility.getUtility(new int[]{i,actions[1],k});
  		  	  totalUtil[i] += util[0]; 
	      }
		  totalUtil[i] = totalUtil[i]/12;
	  }
	  return totalUtil; 
  }
  
  protected void printGame(int[] actions) {
	  if (turn==1) 
		  System.out.println("  "+Array.arrayToNakedString(densities));
	  String beach = "";
	  for (int i = 0; i<numActions; i++) {
		  if (actions[0]==i)
			  beach = beach + "A";
		  else beach = beach + " ";
	  	  if (actions[1]==i)
			  beach = beach + "B";
	  	  else beach = beach + " ";
	  	  if (actions[2]==i)
			  beach = beach + "C";
	  	  else beach = beach + " ";
	  }
	  int[] util = utility.getUtility(actions);
	  System.out.println(beach + Array.arrayToNakedString(util));
  }
  
  protected int bestPayoffs(int[] util) {
	  int maxV = 0;
	  int maxA = 0;
	  for (int i = 0; i<util.length; i++) {
		  if (maxV > util[i]) {
			  maxV = util[i];
			  maxA = i;
		  }
		  if (maxV == util[i] && (maxV/2)%2==1) {
			  maxV = util[i];
			  maxA = i;
		  }
	  }
	  return maxA;
  }
  
  protected int[] currentPayoffs(int[] actions) {
	  int[] util;
	  int[] totalUtil = new int[numActions];
	  for (int i = 0; i<numActions; i++) {
	  		  util = utility.getUtility(new int[]{i,actions[1],actions[2]});
  		  	  totalUtil[i] = util[0]; 
	  }
	  return totalUtil; 
  }
  
}
