package analysis;
import java.util.Random;


public class GameGenerator implements GameGeneratorInterface{
  public GameGenerator(int maxValue, boolean deterministicPairs,
      boolean symmetricDistribution,
      UtilityDistribution utilityDistributionType) {
    this.maxValue = maxValue;
    this.deterministicPairs = deterministicPairs;
    this.symmetricDistribution = symmetricDistribution;
    this.utilityDistributionType = utilityDistributionType;
  }

  final int maxValue;
  final boolean deterministicPairs;
  final boolean symmetricDistribution;
  
  final UtilityDistribution utilityDistributionType;

  
  public GameGenerator(int maxValue,boolean deterministicPairs, boolean symmetricDistribution){
    this(maxValue,deterministicPairs,symmetricDistribution,new StickBreakingUtilityDistribution());
  }

  public GameGenerator(int maxValue,boolean deterministicPairs){
    this(maxValue,deterministicPairs,true);
  }


  public void setPermutation(ThreePlayerGame g, int[] a, int[] u,int i, int j, int k){
    int[] utility = g.getUtility(a[i],a[j],a[k]);
    
    utility[0]=u[i];
    utility[1]=u[j];
    utility[2]=u[k];
  }
  
  public void setJointActionSymmetrically(ThreePlayerGame g, int[] a, int[] u){
    setPermutation(g,a,u,0,1,2);
    setPermutation(g,a,u,0,2,1);
    setPermutation(g,a,u,1,0,2);
    setPermutation(g,a,u,1,2,0);
    setPermutation(g,a,u,2,0,1);
    setPermutation(g,a,u,2,1,0);
  }
  
  public void shuffle(int[] array, Random r){
    for(int i=0;i<array.length;i++){
      int target = r.nextInt(array.length-i)+i;
      if (target>i){
        final int temp=array[target];
        array[target]=array[i];
        array[i]=temp;
      }
    }
  }
  
  
  
  public void generatePairedActions(ThreePlayerGame g, Random r, int NUM_ACTIONS){
    int[] actions = new int[3];
    int[] values = new int[3];
    for(int i=0;i<NUM_ACTIONS;i++){
      for(int j=0;j<NUM_ACTIONS;j++){
        if (j==i){
          continue;
        }
        actions[0]=i;
        actions[1]=j;
        actions[2]=j;
        generateRandomUtilities(values,r);
        values[0] = values[r.nextInt(3)];
        int remaining = maxValue - values[0];
        if ((remaining & 1)==1){
          remaining--;
          values[0]++;
        }
        values[1]=remaining/2;
        values[2]=remaining/2;
        setJointActionSymmetrically(g,actions,values);
        
      }
    }

  }
  

  
  public void generatePairedDeterministically(ThreePlayerGame g, int NUM_ACTIONS){
    int[] actions = new int[3];
    int[] values = new int[3];
    
    for(int i=0;i<NUM_ACTIONS;i++){
      for(int j=0;j<NUM_ACTIONS;j++){
        if (j==i){
          continue;
        }
        actions[0]=i;
        actions[1]=j;
        actions[2]=j;
        values[0]=maxValue/2;
        values[1]=maxValue/4;
        values[2]=maxValue/4;
        setJointActionSymmetrically(g,actions,values);
      }
    }

  }
  
  
  public void generateRandomUtilities(int[] values, Random r){
    utilityDistributionType.generateUtilities(values, maxValue, r);

  }
  
  
  
  /**
   * Generate utilities when all players choose unique actions. If symmetricDistribution is true,
   * then the result of generateRandomUtilities is shuffled after it is generated.
   * @param g
   * @param r
   * @param numActions
   */
  public void generateUniqueActions(ThreePlayerGame g, Random r, int numActions){
    int[] actions = new int[3];
    int[] values = new int[3];
    
    for(int i=0;i<numActions-2;i++){
      for(int j=i+1;j<numActions-1;j++){
        for(int k=j+1;k<numActions;k++){
          actions[0]=i;
          actions[1]=j;
          actions[2]=k;
          generateRandomUtilities(values,r);
          if (symmetricDistribution){
            shuffle(values,r);
          }
          setJointActionSymmetrically(g,actions,values);
        }
      }
    }
  }
  
  /**
   * Generates equal utilities for all the players when the actions are identical.
   * @param g
   * @param numActions
   */
  public void generateSymmetricActions(ThreePlayerGame g, int numActions){
    final int AVERAGE_VALUE = maxValue/3;
    for(int i=0;i<numActions;i++){
      g.data[i][i][i][0]=AVERAGE_VALUE;
      g.data[i][i][i][1]=AVERAGE_VALUE;
      g.data[i][i][i][2]=AVERAGE_VALUE;
    }
  }

  
  /**
   * Generate a symmetric game, where:
   * 1. The game is invariant under a permutation of the players.
   * 2.
   */
  public ThreePlayerGame generateGame(Random r, int numActions){
    ThreePlayerGame g = new ThreePlayerGame(numActions);
    
    generateUniqueActions(g,r,numActions);
    
    generateSymmetricActions(g,numActions);
    if (deterministicPairs){
      generatePairedDeterministically(g,numActions);
    } else {
      generatePairedActions(g,r,numActions);
    }
    return g;
  }
}
