package com.yahoo.lsg.lifelong;

import game.ConstantGameGenerator;
import game.Game;
import game.GameGenerator;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import com.yahoo.lsg.lifelong.observer.LifelongObserver;
import com.yahoo.lsg.lifelong.observer.UtilityObserver;

import lemonade.ScaledLemonadeUtility;


/**
 * Runs a lifelong learning tournament. The tournament consists of several lifetimes. During each lifetime, the players maintain state.
 * They are notified at the beginning of each game they participate in what the utilities are, but not who they are playing against.
 * During the tournament, the total utility of each player and the total utility squared are maintained, such that the mean and the 
 * standard error of the performance of each bot can be calculated.
 * 
 * @author maz
 *
 */

public class LifelongLearningTournament {
  final LifelongObserver observer;
  
  
  
  private static Constructor<? extends LifelongObserver> getConstructor(Class<? extends LifelongObserver> observerClass) throws SecurityException, NoSuchMethodException{
    Class<?> infoClass = GameInformation.class;
    Class<?> identityClass = new PlayerIdentity[0].getClass();
    return observerClass.getConstructor(infoClass,identityClass);
  }
  /**
   * 
   * @param info the parameters of the game being played.
   * @param generator the generator for games to be played.
   * @param playerClasses the classes of the players: must extend com.yahoo.lifelong.LifelongStrategy
   * @param playerOptions the options of the players
   * @param seedGenerator a generator for shuffling players and giving seeds to players
   * @throws SecurityException if the constructor of a player class is not accessible.
   * @throws ClassNotFoundException if the class of a player is not found.
   * @throws NoSuchMethodException if there is no constructor method for the player class.
   * @throws InvocationTargetException 
   * @throws IllegalAccessException 
   * @throws InstantiationException 
   * @throws IllegalArgumentException 
   */
  public LifelongLearningTournament(GameInformation info, GameGenerator generator,
      String[] playerClasses, String[] playerOptions, Random seedGenerator, Class<? extends LifelongObserver> observerClass) throws SecurityException, ClassNotFoundException, NoSuchMethodException, IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException{
    super();
    this.info = info;
    this.generator = generator;
    this.seedGenerator = seedGenerator;
    players = new PlayerIdentity[playerClasses.length];
    for(int i=0;i<players.length;i++){
      players[i]=new PlayerIdentity(playerClasses[i],playerOptions[i]);
    }
    game = null;
    playersRemaining = new ArrayList<PlayerIdentity>();
    
    
    
    Constructor<? extends LifelongObserver> observerConstructor = getConstructor(observerClass);
    
    this.observer = observerConstructor.newInstance(info,players);
    
  }

  
  /**
   * The game currently being played.
   */
  private Game game;
  
  /**
   * The options for the competition.
   */
  private final GameInformation info;
  
  /**
   * The game generator.
   */
  private final GameGenerator generator;
  
  /**
   * The player records for the game. Includes the strategy
   * (if initialized) and a summary of the utility over 
   * all the lifetimes.
   */
  private final PlayerIdentity[] players;
  
  /**
   * A seed generator. Used to generate seeds for the players,
   * and to shuffle the players before each game.
   */
  private final Random seedGenerator;
  
  /**
   * A list of players who have not played the current game.
   * Initially, it is empty.
   */
  private final List<PlayerIdentity> playersRemaining;
  
  
  /**
   * Choose a new game,
   * and reset players remaining.
   * 
   * Used if there are not enough players left who have not played the
   * current game.
   */
  private void chooseNewGame(){
    game = generator.getNextGame();
    playersRemaining.clear();
    for(PlayerIdentity record:players){
      playersRemaining.add(record);
    }
  }
  
  
  /**
   * Gets a remaining player uniformly at random, removing him from the list.
   * If there are no players remaining, throws a RuntimeException.
   * 
   * This code is not thread safe.
   * @return
   */
  private PlayerIdentity getRemainingPlayer(){
    if (playersRemaining.isEmpty()){
      throw new RuntimeException("Tried to get a remaining player, but there are none remaining");
    }
    int playerIndex = seedGenerator.nextInt(playersRemaining.size());
    return playersRemaining.remove(playerIndex);
  }
  
  /**
   * Get a matchup. This code is not thread safe.
   * @return
   */
  public LifelongMatchup getLifelongMatchup(){
    if (game==null || (playersRemaining.size()<game.getNumPlayers())){
      chooseNewGame();
      assert(playersRemaining.size()>=game.getNumPlayers());
    }
    final PlayerIdentity[] records = new PlayerIdentity[game.getNumPlayers()];
    for(int i=0;i<records.length;i++){
      records[i]=getRemainingPlayer();
    }
    return new LifelongMatchup(records);
  }
  
  
  /**
   * Initialize all the players' strategy objects, and generate the next seed.
   */
  public void initializeLifetime() {
    for(int i=0;i<players.length;i++){
      long nextSeed = seedGenerator.nextLong();
      try {
        players[i].initialize(info,nextSeed);
      } catch (Exception e) {
        throw new RuntimeException("Error constructing player "+i+".",e);
      }
    }
    
    playersRemaining.clear();
  }
  
  /**
   * Play a single match.
   */
  public void playMatch(){
    final LifelongMatchup matchup = getLifelongMatchup();
    matchup.playMatch(game, info.numRounds, observer);
  }
  
  public void playLifetime(){
    initializeLifetime();
    
    for(int i=0;i<info.numMatches;i++){
      playMatch();
    }
    for(int i=0;i<players.length;i++){
      players[i].endLifetime();
    }
    observer.endLifetime();
  }
  
  
  
  
  public void run(final int repeat){
    for(int i=0;i<repeat;i++){
      playLifetime();
    }
  }
  
  
  
  public void display(){
    observer.display();
  }
  
  
  public static void showUsage(){
    System.err.println("LifelongLearningTournament [-seed <SEED> | -matchtype <MATCHTYPE> | -rounds <ROUNDS> | -epochs <EPOCHS>| -repeats <REPEATS>| -generator <GENERATOR>| -actions <ACTIONS>]* -players [<CLASSNAME> <OPTIONS>]*");
    System.exit(0);
  }
  public static void main(String[] args) throws ClassNotFoundException, SecurityException, NoSuchMethodException, IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException{
    if (args.length<4){
      showUsage();
    }
    
    long seed=0;
    int numActions=12;
    int rounds = 100;
    int repeats = 1000;
    int matches = 20;
    String observerClassName = UtilityObserver.class.getName();
    int playerArgIndex=args.length;
    boolean defaultOptions = false;
    //String matchType = "normal";
    String generator = null;
    for(int i=0;i<args.length;i+=2){
      String arg = args[i];
      String fieldStr = args[i+1];
      if (arg.equals("-seed")){
        long field = Long.parseLong(fieldStr);
        seed = field;
      } else if (arg.equals("-observer")){
        observerClassName = fieldStr;
      } else if (arg.equals("-generator")){
        generator = fieldStr;
        
        
      }else if (arg.equals("-actions")){
        long field = Long.parseLong(fieldStr);
        if (100L<field){
          System.err.println("Number of actions needs to be less than 100");
          showUsage();
        }
        numActions = (int)field;
      } else if (arg.equals("-rounds")){
        long field = Long.parseLong(fieldStr);
        if ((long)(Integer.MAX_VALUE)<field){
          System.err.println("Number of rounds needs to be less than 2^31-1");
          showUsage();
        }
        rounds = (int)field;
      } else if (arg.equals("-matches")){
        long field = Long.parseLong(fieldStr);
        if ((long)(Integer.MAX_VALUE)<field){
          System.err.println("Number of matches needs to be less than 2^31-1");
          showUsage();
        }
        matches = (int)field;
      } else if (arg.equals("-repeats")){
        long field = Long.parseLong(fieldStr);
        if ((long)(Integer.MAX_VALUE)<field){
          System.err.println("Number of repeats needs to be less than 2^31-1");
          showUsage();
        }
        repeats = (int)field;
      } else if (arg.equals("-defaultOptions")){
        defaultOptions = true;
      } else if (arg.equals("-players")){
        playerArgIndex=i+1;
        break;
      } else {
        System.err.println("Unknown option:"+arg);
        showUsage();
        
      }
    }

    Random seedGenerator = new Random(seed);
    ScaledLemonadeUtility utility = new ScaledLemonadeUtility(numActions);
    GameGenerator g = new ConstantGameGenerator(new Game(new int[]{numActions,numActions,numActions},utility));

    if (generator!=null){
      long generatorSeed = seedGenerator.nextLong();
      Class<?> generatorClass = Class.forName(generator);
      Constructor<?> constructor = generatorClass.getConstructor(long.class);
      g = (GameGenerator)constructor.newInstance(generatorSeed);      
    }
    
    
    final String[] strategies;
    final String[] options;
    int playerIndex = 0;
    if (defaultOptions){
      strategies = new String[(args.length-playerArgIndex)];
      options = new String[strategies.length];
    
      for(int i=playerArgIndex;i<args.length;i++){
        strategies[playerIndex]=args[i];
        options[playerIndex]="";
        playerIndex++;
      } 
      
    } else {
      strategies = new String[(args.length-playerArgIndex)/2];
      options = new String[strategies.length];
    
      for(int i=playerArgIndex;i<args.length;i+=2){
        strategies[playerIndex]=args[i];
        options[playerIndex]=args[i+1];
        playerIndex++;
      } 
    }
    GameInformation info = new GameInformation(playerIndex,rounds,matches);
    
    @SuppressWarnings("unchecked")
    final Class<? extends LifelongObserver> observerClass = (Class<? extends LifelongObserver>) Class.forName(observerClassName);
    
    LifelongLearningTournament tournament = new LifelongLearningTournament(info, g, strategies, options, seedGenerator, observerClass);
    tournament.run(repeats);
    tournament.display();
  }
}
