package com.yahoo.lsg.sundisplay;
import game.Game;
import game.Utility;

import java.awt.BasicStroke;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Point;
import java.awt.Stroke;
import java.util.Vector;

import lemonade.Action;
import lemonade.GeneralizedLemonadeUtility;
import lemonade.IslandLogic;


/**
 * A view for a human to play the game. The meat of the
 * state of the data is in the HumanModel to which this points.
 * @author maz
 *
 */
public class SunDisplay extends Canvas implements SunModelListener{
  /**
   * 
   */
  private static final long serialVersionUID = 1L;
  private Game game;
  private IslandLogic logic;
  
  private SunModelData model;
  
  Image buffer;
  
  
  private Color[] colors;
  private Color[] fadedColors;
  
  private static final Color[] DEFAULT_COLORS = new Color[]{new Color(255,0,255),new Color(0,196,0),Color.red,Color.blue,new Color(0,234,0)};
  
  private int MAX_UTILITY;
  
  private String[] names;
  
  /**
   * The sky is a lighter blue than the ocean.
   */
  private static final Color SKY_COLOR = new Color(196,196,255);
  
  /**
   * The ocean is a dark blue
   */
  private static final Color OCEAN_COLOR = new Color(0,0,255);

  /**
   * The beach is a light yellow
   */
  private static final Color BEACH_COLOR = new Color(255,255,96);
  
  /**
   * The jungle (island interior) is a dark green
   */
  private static final Color JUNGLE_COLOR = new Color(0,128,0);
  
  static final Color VACANT_SPOT_COLOR = Color.black;
  
  
  public static int getMaxUtility(Game g){
    Utility u = g.getUtility();
    int maxSoFar = u.getUtility(new int[]{0,0,0})[0];
    for(int i=0;i<g.getNumActions(0);i++){
      for(int j=0;j<g.getNumActions(1);j++){
        for(int k=0;k<g.getNumActions(2);k++){
          int[] util = u.getUtility(new int[]{i,j,k});
          for(int m=0;m<util.length;m++){
            if (util[m]>maxSoFar){
              maxSoFar = util[m];
            }
          }
        }
      }
    }
    return maxSoFar;
  }
  
  public SunDisplay(Game g){
    this(g,new String[]{"you","",""});
  }
  
  
  
  public SunDisplay(Game g, String[] names){
    assert(g.getNumPlayers()==3);
    
    // Maximum utility should not be zero or bad things happen.
    MAX_UTILITY = Math.max(getMaxUtility(g),1);
    model = new SunModelData(g.getNumPlayers());
    this.game=g;
    this.logic = new IslandLogic(g.getNumActions());
    colors = new Color[g.getNumPlayers()];
    System.arraycopy(DEFAULT_COLORS, 0, colors, 0, colors.length);
    fadedColors = new Color[g.getNumPlayers()];
    for(int i=0;i<colors.length;i++){
      int red = colors[i].getRed();
      int green = colors[i].getGreen();
      int blue = colors[i].getBlue();
      int fadedRed = red*2/3;
      int fadedGreen = green*2/3;
      int fadedBlue = blue*2/3;
      fadedColors[i]=new Color(fadedRed,fadedGreen,fadedBlue);
    }
    this.names = new String[]{"Purple ("+names[0]+")","Green ("+names[1]+")", "Red ("+names[2]+")"};
  }
  
  /**
   * The ideal size of the canvas.
   */
  private static final Dimension size = new Dimension(640,680);
  
  @Override
  public Dimension getMaximumSize() {
    return size;
  }



  @Override
  public Dimension getMinimumSize() {
    return size;
  }



  @Override
  public Dimension getPreferredSize() {
    return size;
  }

  private int width;
  private int height;
  private int minDim;
  private int radius;
  private int positionRadius;
  private int pieceRadius;
  private int centerx;
  private int centery;
  
  private int skyHeight;

  private void evaluateResize(){
    width = getWidth();
    height = getHeight();
    
    skyHeight = 200;
    minDim = Math.min(width, height-skyHeight);
    radius = (minDim*2)/5;
    positionRadius = radius/10;
    centerx = width/2;
    centery = (height-skyHeight)/2 + skyHeight;
    pieceRadius = positionRadius/2;
    if (buffer!=null){
      int imgHeight = buffer.getHeight(this);
      int imgWidth = buffer.getWidth(this);
      if (imgHeight!=-1 && imgHeight!=height){
        buffer=createImage(width,height);
      } else if (imgWidth!=-1 && imgWidth!=width){
        buffer=createImage(width,height);
      }
    } else {
      buffer = createImage(width,height);
    }
  }
  
  
  /**
   * Get the location of the point (in terms of the positions in the game). -1 indicates no circle was 
   * clicked on.
   * @param p the query point
   * @return the position (0 to islandSize-1), or -1 if a position was not clicked.
   */
  public int getPosition(Point p){
    int x = p.x;
    int y = p.y;
    int positionRadiusSquared = positionRadius*positionRadius;
    Point center = new Point();
    for(int i=0;i<game.getNumActions();i++){
      getPositionPoint(i,center);
      int deltaX = (x-center.x);
      int deltaY = (y-center.y);
      int distanceSquared = (deltaX*deltaX)+(deltaY*deltaY);
      if (distanceSquared<=positionRadiusSquared){
        return i;
      }
    }
    return -1;
  }

  private void getPositionPoint(int position, Point p){
    double angle = Math.PI*2*position/game.getNumActions();
    p.x = centerx+(int)(radius * Math.sin(angle));
    p.y = centery+(int)(radius * Math.cos(angle));
    
  }

  public void getPositionForPlayer(int position,int player, Point p){
    this.getPositionPoint(position,p);
    double angle = Math.PI*2*player/game.getNumPlayers();
    p.x+=(int)(pieceRadius * Math.sin(angle));
    p.y+=(int)(pieceRadius * Math.cos(angle));
    p.x-=pieceRadius;
    p.y-=pieceRadius;

  }
  
  
  
  
  /**
   * Repaints the display. By the contract of HumanModelListener, since this is called after
   * handleChange, this is not allowed to obtain any locks.
   */
  @Override
  public void paint(Graphics g){
    update(g);
  }
  
  @Override
  public void update(Graphics g){
    evaluateResize();
    if (buffer!=null){
      Graphics g2 = buffer.getGraphics();
      if (g2!=null){
        paintOn(g2);
        g.drawImage(buffer, 0, 0, this);
        return;
      }
    }
    System.err.println("Buffered image was not ready");
    // If the buffer is not ready, paint directly.
    paintOn(g);
  }
  
  public void paintOn(Graphics g){
    int pieceDiameter = pieceRadius*2;
    int positionDiameter = positionRadius*2;
    
    int beachOffset = ((positionRadius*5)/4);

    
    boolean isNight = model.getTimeOfDay()<0.5;
    if (isNight){
      g.setColor(Color.BLACK);
    } else {
      g.setColor(SunDisplay.SKY_COLOR);
    }
    g.fillRect(0,0,width,skyHeight);

    int sunRadius = 20;
    double sunPathCenterX = width/2;
    double sunPathCenterY = skyHeight+sunRadius;
    double sunPathRadius = skyHeight;
    
    double timeOfHalfDay = (isNight) ? (model.getTimeOfDay()*2.0) : (model.getTimeOfDay()*2.0-1.0);
    double sunAngle = Math.PI * (1.0- timeOfHalfDay);
    int sunCenterX = (int)(sunPathCenterX+Math.cos(sunAngle) *sunPathRadius); 
    int sunCenterY = (int)(sunPathCenterY-Math.sin(sunAngle)*sunPathRadius);
    
    if (isNight){
      g.setColor(Color.lightGray);
      g.fillOval(sunCenterX-sunRadius, sunCenterY-sunRadius, sunRadius*2, sunRadius*2);
    } else {
      g.setColor(Color.YELLOW);
    g.fillOval(sunCenterX-sunRadius, sunCenterY-sunRadius, sunRadius*2, sunRadius*2);
    int rayLength = sunRadius * 2;
    for(int i=0;i<12;i++){
      double rayAngle = (Math.PI * 2 * i)/12;
      int xPos = (int)(sunCenterX+Math.cos(rayAngle) * rayLength);
      int yPos = (int)(sunCenterY+Math.sin(rayAngle) * rayLength);
      g.drawLine(xPos,yPos,sunCenterX,sunCenterY);
    }
    }
    //System.out.println("Sun:"+sunCenterX+","+sunCenterY +" timeOfDay:"+model.getTimeOfDay());
    
    
    g.setColor(SunDisplay.OCEAN_COLOR);
    g.fillRect(0,skyHeight,width,height-skyHeight);
    
    int borderRespected = height/24;

    int beachRadius = radius + beachOffset;
    int beachDiameter = beachRadius*2;
    int beachTLX = centerx-beachRadius;
    int beachTLY = centery-beachRadius;
    
    g.setColor(SunDisplay.BEACH_COLOR);
    g.fillOval(beachTLX, beachTLY, beachDiameter, beachDiameter);
    
    int jungleRadius = radius - beachOffset;
    int jungleDiameter = jungleRadius*2;
    int jungleTLX = centerx-jungleRadius;
    int jungleTLY = centery-jungleRadius;
    g.setColor(SunDisplay.JUNGLE_COLOR);
    g.fillOval(jungleTLX, jungleTLY, jungleDiameter, jungleDiameter);
    
    
    Point p = new Point();
    for(int i=0;i<game.getNumActions();i++){
      g.setColor(SunDisplay.VACANT_SPOT_COLOR);
      this.getPositionPoint(i, p);
      int cornerPosX = p.x-positionRadius;
      int cornerPosY = p.y-positionRadius;
      g.fillOval(cornerPosX, cornerPosY, positionDiameter, positionDiameter);
      //g.setColor(Color.white);
      //g.drawString(""+i, p.x, p.y);
    }
    //System.out.println("SunDisplay:model.positions="+Arrays.toString(model.positions));
    //System.out.println("SunDisplay:model.lastActions="+Arrays.toString(model.lastActions));
    
    Point playerPos = new Point();
    Point oldPlayerPos = new Point();
    Point nextPlayerPos = new Point();
    for(int i=0;i<model.positions.length;i++){
      if (model.isVisible(i)){
        if (model.isFaded(i)){
          g.setColor(fadedColors[i]);
        } else {
          g.setColor(colors[i]);
        }

        if (isNight && (model.lastActions[i]!=model.positions[i])){
          getPositionForPlayer(model.lastActions[i],i,oldPlayerPos);
          getPositionForPlayer(model.positions[i],i,nextPlayerPos);
          playerPos.x = (int)((1.0-timeOfHalfDay)*oldPlayerPos.x+timeOfHalfDay*nextPlayerPos.x);
          playerPos.y = (int)((1.0-timeOfHalfDay)*oldPlayerPos.y+timeOfHalfDay*nextPlayerPos.y);
        } else {
          getPositionForPlayer(model.positions[i],i,playerPos);
        }        
        /*
        this.getPositionPoint(model.positions[i],p);
        double angle = Math.PI*2*i/game.getNumPlayers();
        int pieceX = p.x+(int)(pieceRadius * Math.sin(angle));
        int pieceY = p.y+(int)(pieceRadius * Math.cos(angle));
        int cornerPieceX = pieceX-pieceRadius;
        int cornerPieceY = pieceY-pieceRadius;
        */
        g.fillOval(playerPos.x, playerPos.y, pieceDiameter, pieceDiameter);
      }
    }
    if (model.allVisible() && model.utilitiesVisible){
      Vector<Action> sortedPositions = logic.getPositionsSorted(model.positions);
      Graphics2D g2 = (Graphics2D)g;

      Stroke s = g2.getStroke();

      for(int i=0;i<sortedPositions.size();i++){
        Action a = sortedPositions.get(i);
        int piecesOnSpotBefore=0;
        for(int j=0;j<a.player;j++){
          if (model.positions[j]==a.position){
            piecesOnSpotBefore++;
          }
        }
        int posDegrees = 360*a.position/game.getNumActions()-90;
        int arcClockwise = -180*logic.getDistanceRight(sortedPositions, i)/game.getNumActions();
        int arcCounterClockwise = 180*logic.getDistanceLeft(sortedPositions, i)/game.getNumActions();
        int numOnTop = logic.numberOfPlayersOnTop(sortedPositions, i);
        float lineWidth = ((float)pieceRadius)/numOnTop;
        int lineRadius = beachRadius-1+(int)(lineWidth*(0.5+piecesOnSpotBefore));
        g2.setColor(colors[a.player]);
        BasicStroke s2 = new BasicStroke(lineWidth,BasicStroke.CAP_SQUARE,BasicStroke.JOIN_ROUND);
        g2.setStroke(s2);
        g.drawArc(centerx-lineRadius, centery-lineRadius, lineRadius*2, lineRadius*2, posDegrees,arcClockwise);
        g.drawArc(centerx-lineRadius, centery-lineRadius, lineRadius*2, lineRadius*2, posDegrees,arcCounterClockwise);

      }
      Utility u = game.getUtility();
      if (u instanceof GeneralizedLemonadeUtility){
        GeneralizedLemonadeUtility glu = (GeneralizedLemonadeUtility )u;
        int[] density = glu.getIslandDensity();
        FontMetrics fm = g.getFontMetrics();
        
        for(int i=0;i<game.getNumActions();i++){
          this.getPositionPoint(i, p);
          int numCustomers = density[i];
          String current = ""+numCustomers;
          int cornerPosX = p.x-(fm.charsWidth(current.toCharArray(), 0, current.length())/2);
          int cornerPosY = p.y+(fm.getAscent()/2);
          
          //int cornerPosX = p.x-positionRadius;
          //int cornerPosY = p.y-positionRadius;
          g.setColor(Color.white);
          g.drawString(current, cornerPosX, cornerPosY);
          
          int customerRadius = 3;
          
          int[] numCustomersPerPlayer=new int[3];
          int minDistance = glu.getDistance(i, model.positions[0]);
          int numWinners = 1;
          int firstWinner = 0;
          for(int j=1;j<numCustomersPerPlayer.length;j++){
            int currentDistance = glu.getDistance(i, model.positions[j]);
            if (currentDistance==minDistance){
              numWinners++;
            } else if (currentDistance<minDistance){
              minDistance=currentDistance;
              numWinners = 1;
              firstWinner = j;
            }
          }
          
          
          int currentWinner = firstWinner;
          int numToWin = numCustomers/numWinners;
          
          try{
          for(int j=0;j<numCustomers;j++){
            double angle = (Math.PI * 2 * j)/numCustomers;
            int customerX = p.x+(int)(Math.cos(angle)*positionRadius);
            int customerY = p.y+(int)(Math.sin(angle)*positionRadius);
            g.setColor(colors[currentWinner]);
            g.fillOval(customerX-customerRadius, customerY-customerRadius, 2*customerRadius, 2 * customerRadius);
            numToWin--;
            if (numToWin==0){
              currentWinner++;
              while(currentWinner<3 && (glu.getDistance(i,model.positions[currentWinner])>minDistance)){
                currentWinner++;
              }
              numToWin=numCustomers/numWinners;
            }
          }
          } catch (ArrayIndexOutOfBoundsException e){
            e.printStackTrace();
          }
        }
      }
      
      int legendHeight = (height-skyHeight)/6;
      double utilScale = ((double)legendHeight)/MAX_UTILITY;
      legendHeight = (int)(utilScale* MAX_UTILITY);
      int legendWidth = width/6;
      int legendx = width-(borderRespected+legendWidth);
      // The lower part of the legend.
      int legendy = height-borderRespected;
      int legendBorder = 5;
      int strokeSize = legendWidth*2/(3*game.getNumPlayers());
      int spacing = (legendWidth-(strokeSize+(legendBorder*2)))/(game.getNumPlayers()-1);
      int[] utility = game.getUtility().getUtility(model.positions);
      
      g2.setStroke(s);
      g2.setColor(Color.white);
      
      
      if (utilScale*2>5.0){
      for(int i=0;i<=MAX_UTILITY;i+=2){
        g2.drawLine(legendx, legendy-(int)(utilScale*i), legendx+legendWidth, legendy-(int)(utilScale*i));
      }
      }
      
      
      g2.setStroke(new BasicStroke(strokeSize,BasicStroke.CAP_BUTT,BasicStroke.JOIN_MITER));
      for(int i=0;i<utility.length;i++){
        int xCoord = legendx+(spacing*i)+(strokeSize/2)+legendBorder;
        int lineTop = legendy-(int)(utilScale*utility[i]);
        int lineBottom = legendy;
        //System.out.println(lineTop);
        g2.setColor(colors[i]);
        g2.drawLine(xCoord,lineTop,xCoord,lineBottom);
      }
      g2.setStroke(new BasicStroke(3,BasicStroke.CAP_BUTT,BasicStroke.JOIN_MITER));
      g2.setColor(Color.white);
      //g2.drawLine(legendx, legendy, legendx, legendy-legendHeight);
      g2.drawRect(legendx, legendy-legendHeight, legendWidth, legendHeight);
      g2.setStroke(s);
      
    }
    String modelMessage = model.message;
    g.setColor(Color.white);
    if (modelMessage!=null){
      int stringWidth = g.getFontMetrics().charsWidth(modelMessage.toCharArray(), 0, modelMessage.length());
      g.drawString(model.message,centerx-(stringWidth/2),centery);
    }
    FontMetrics fm =g.getFontMetrics();
    int maxNameWidth = 0;
    for(int i=0;i<names.length;i++){
      int nameWidth = fm.charsWidth(names[i].toCharArray(), 0, names[i].length());
      maxNameWidth = Math.max(nameWidth, maxNameWidth);
      int y = borderRespected+(i*g.getFontMetrics().getHeight())+g.getFontMetrics().getAscent();
      int x = borderRespected;
      g.drawString(names[i],x,y);
    }
    String biggestScore=""+(20*MAX_UTILITY);
    int scoreLegendWidth = fm.charsWidth(biggestScore.toCharArray(), 0, biggestScore.length());
    for(int i=0;i<model.scores.length;i++){
      String scoreString = ""+model.scores[i];
      int stringWidth = fm.charsWidth(scoreString.toCharArray(), 0, scoreString.length());
      int y = borderRespected+(i*g.getFontMetrics().getHeight())+g.getFontMetrics().getAscent();
      int x = borderRespected+(scoreLegendWidth+maxNameWidth-stringWidth);
      g.drawString(scoreString,x,y);
    }
    if (model.currentGameIndex!=null){
    int gameIndexX=borderRespected;
    int gameIndexY=height-borderRespected;
    
    g.drawString(""+model.currentGameIndex, gameIndexX, gameIndexY);
    }
  }



  public void handleChange(SunModel model) {
    model.getModel(this.model);
    repaint();
  }
}
